#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import pickle
import queue
import threading
from typing import TYPE_CHECKING

from fim.graph.abc_property_graph import ABCPropertyGraph

from fabric.actor.core.apis.i_actor import ActorType
from fabric.actor.core.apis.i_delegation import IDelegation
from fabric.actor.core.common.exceptions import BrokerException
from fabric.actor.core.delegation.broker_delegation_factory import BrokerDelegationFactory
from fabric.actor.core.delegation.delegation_factory import DelegationFactory
from fabric.actor.core.kernel.slice_factory import SliceFactory
from fabric.actor.core.manage.broker_management_object import BrokerManagementObject
from fabric.actor.core.manage.kafka.services.kafka_broker_service import KafkaBrokerService
from fabric.actor.core.proxies.kafka.services.broker_service import BrokerService
from fabric.actor.core.registry.actor_registry import ActorRegistrySingleton
from fabric.actor.core.util.resource_data import ResourceData
from fabric.actor.core.util.id import ID
from fabric.actor.core.apis.i_broker_reservation import IBrokerReservation
from fabric.actor.core.apis.i_broker import IBroker
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.core.actor import Actor
from fabric.actor.core.registry.peer_registry import PeerRegistry
from fabric.actor.core.time.actor_clock import ActorClock
from fabric.actor.core.util.reservation_set import ReservationSet
from fabric.actor.security.auth_token import AuthToken

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_broker_proxy import IBrokerProxy
    from fabric.actor.core.apis.i_slice import ISlice
    from fabric.actor.core.apis.i_client_reservation import IClientReservation
    from fabric.actor.core.apis.i_client_callback_proxy import IClientCallbackProxy
    from fabric.actor.core.apis.i_reservation import IReservation
    from fabric.actor.core.util.client import Client


class Broker(Actor, IBroker):
    """
    Broker offers the base for all broker actors.
    """
    def __init__(self, *, identity: AuthToken = None, clock: ActorClock = None):
        super().__init__(auth=identity, clock=clock)
        # Recovered reservations that need to obtain tickets (both server and client roles).
        self.ticketing = ReservationSet()
        # Recovered reservations that need to extend tickets (both server and client roles).
        self.extending = ReservationSet()
        # The peer registry.
        self.registry = PeerRegistry()
        # Initialization status.
        self.initialized = False
        self.type = ActorType.Broker

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['recovered']
        del state['wrapper']
        del state['logger']
        del state['clock']
        del state['current_cycle']
        del state['first_tick']
        del state['stopped']
        del state['initialized']
        del state['thread_lock']
        del state['thread']
        del state['timer_queue']
        del state['event_queue']
        del state['reservation_tracker']
        del state['subscription_id']
        del state['actor_main_lock']
        del state['closing']
        del state['message_service']

        del state['ticketing']
        del state['extending']
        del state['registry']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

        self.recovered = False
        self.wrapper = None
        self.logger = None
        self.clock = None
        self.current_cycle = -1
        self.first_tick = True
        self.stopped = False
        self.initialized = False
        self.thread = None
        self.thread_lock = threading.Lock()
        self.timer_queue = queue.Queue()
        self.event_queue = queue.Queue()
        self.reservation_tracker = None
        self.subscription_id = None
        self.actor_main_lock = threading.Condition()
        self.closing = ReservationSet()
        self.message_service = None

        self.ticketing = ReservationSet()
        self.extending = ReservationSet()
        self.registry = PeerRegistry()

    def actor_added(self):
        super().actor_added()
        self.registry.actor_added()

    def add_broker(self, *, broker: IBrokerProxy):
        self.registry.add_broker(broker=broker)

    def register_client_slice(self, *, slice_obj: ISlice):
        self.wrapper.register_slice(slice_object=slice_obj)

    def bid(self, *, cycle: int):
        """
        Bids for resources as dictated by the bidding policy for the current cycle.

        @param cycle cycle

        @throws Exception in case of error
        """
        candidates = None
        candidates = self.policy.formulate_bids(cycle=cycle)
        if candidates is not None:
            for reservation in candidates.get_ticketing().values():
                try:
                    self.wrapper.ticket(reservation=reservation, destination=self)
                except Exception as e:
                    self.logger.error("unexpected ticket failure for #{} e: {}".format(
                        reservation.get_reservation_id(), e))

            for reservation in candidates.get_extending().values():
                try:
                    self.wrapper.extend_ticket(reservation=reservation)
                except Exception as e:
                    self.logger.error("unexpected extend failure for #{}".format(reservation.get_reservation_id()))

    def claim_delegation_client(self, *, delegation_id: str = None, slice_object: ISlice = None,
                                broker: IBrokerProxy = None, id_token: str = None) -> IDelegation:
        if delegation_id is None:
            raise BrokerException(Constants.invalid_argument)

        if broker is None:
            broker = self.get_default_broker()

        if slice_object is None:
            slice_object = self.get_default_slice()
            if slice_object is None:
                slice_object = SliceFactory.create(slice_id=ID(), name=self.identity.get_name())
                slice_object.set_owner(owner=self.identity)
                slice_object.set_inventory(value=True)

        delegation = BrokerDelegationFactory.create(did=delegation_id, slice_id=slice_object.get_slice_id(),
                                                    broker=broker)
        delegation.set_exported(value=True)
        delegation.set_slice_object(slice_object=slice_object)

        self.wrapper.delegate(delegation=delegation, destination=self, id_token=id_token)
        return delegation

    def reclaim_delegation_client(self, *, delegation_id: str = None, slice_object: ISlice = None,
                                  broker: IBrokerProxy = None, id_token: str = None) -> IDelegation:
        if delegation_id is None:
            raise BrokerException(Constants.invalid_argument)

        if broker is None:
            broker = self.get_default_broker()

        if slice_object is None:
            slice_object = self.get_default_slice()
            if slice_object is None:
                slice_object = SliceFactory.create(slice_id=ID(), name=self.identity.get_name())
                slice_object.set_owner(owner=self.identity)
                slice_object.set_inventory(value=True)

        delegation = BrokerDelegationFactory.create(did=delegation_id, slice_id=slice_object.get_slice_id(),
                                                    broker=broker)
        delegation.set_slice_object(slice_object=slice_object)
        delegation.set_exported(value=True)

        callback = ActorRegistrySingleton.get().get_callback(protocol=Constants.protocol_kafka,
                                                             actor_name=self.get_name())
        if callback is None:
            raise BrokerException("Unsupported")

        delegation.prepare(callback=callback, logger=self.logger)
        delegation.validate_outgoing()

        self.wrapper.reclaim_delegation_request(delegation=delegation, caller=broker, callback=callback,
                                                id_token=id_token)

        return delegation

    def claim_delegation(self, *, delegation: IDelegation, callback: IClientCallbackProxy, caller: AuthToken,
                         id_token: str = None):
        if not self.is_recovered() or self.is_stopped():
            raise BrokerException(Constants.invalid_actor_state)
        self.wrapper.claim_delegation_request(delegation=delegation, caller=caller, callback=callback,
                                              id_token=id_token)

    def reclaim_delegation(self, *, delegation: IDelegation, callback: IClientCallbackProxy, caller: AuthToken,
                           id_token: str = None):
        if not self.is_recovered() or self.is_stopped():
            raise BrokerException(Constants.invalid_actor_state)
        self.wrapper.reclaim_delegation_request(delegation=delegation, caller=caller, callback=callback,
                                                id_token=id_token)

    def close_expiring(self, *, cycle: int):
        """
        Closes all expiring reservations.
        @param cycle cycle
        """
        expired = self.policy.get_closing(cycle=cycle)
        if expired is not None:
            self.logger.info("Broker expiring for cycle {} = {}".format(cycle, expired))
            self.close_reservations(reservations=expired)

    def demand(self, *, rid: ID):
        if rid is None:
            raise BrokerException(Constants.invalid_argument)
        reservation = self.get_reservation(rid=rid)
        if reservation is None:
            raise BrokerException("unknown reservation #{}".format(rid))

        self.policy.demand(reservation=reservation)
        reservation.set_policy(policy=self.policy)

    def donate_reservation(self, *, reservation: IClientReservation):
        self.policy.donate_reservation(reservation=reservation)

    def advertise(self, *, delegation: ABCPropertyGraph, client: AuthToken) -> ID:
        slice_obj = SliceFactory.create(slice_id=ID(), name=client.get_name(), data=ResourceData())
        slice_obj.set_owner(owner=client)
        slice_obj.set_broker_client()

        dlg_obj = DelegationFactory.create(did=delegation.get_graph_id(), slice_id=slice_obj.get_slice_id())
        dlg_obj.set_slice_object(slice_object=slice_obj)
        dlg_obj.set_graph(graph=delegation)
        self.wrapper.advertise(delegation=dlg_obj, client=client)
        return dlg_obj.get_delegation_id()

    def extend_ticket_client(self, *, reservation: IClientReservation):
        if not self.recovered:
            self.extending.add(reservation=reservation)
        else:
            self.wrapper.extend_ticket(reservation=reservation)

    def extend_tickets_client(self, *, rset: ReservationSet):
        for reservation in rset.values():
            try:
                if isinstance(reservation, IBrokerReservation):
                    self.extend_ticket_broker(reservation=reservation)
                elif isinstance(reservation, IClientReservation):
                    self.extend_ticket_client(reservation=reservation)
                else:
                    self.logger.warning("Reservation #{} cannot be ticketed".format(reservation.get_reservation_id()))
            except Exception as e:
                self.logger.error("Could not ticket for # {} e: {}".format(reservation.get_reservation_id(), e))

    def extend_ticket_broker(self, *, reservation: IBrokerReservation):
        if not self.recovered:
            self.extending.add(reservation=reservation)
        else:
            self.wrapper.extend_ticket_request(reservation=reservation, caller=reservation.get_client_auth_token(),
                                               compare_sequence_numbers=False)

    def extend_ticket(self, *, reservation: IReservation, caller: AuthToken):
        if not self.recovered or self.is_stopped():
            raise BrokerException(Constants.invalid_actor_state)
        self.wrapper.extend_ticket_request(reservation=reservation, caller=caller, compare_sequence_numbers=True)

    def get_broker(self, *, guid: ID) -> IBrokerProxy:
        return self.registry.get_broker(guid=guid)

    def get_brokers(self) -> list:
        return self.registry.get_brokers()

    def get_default_broker(self) -> IBrokerProxy:
        return self.registry.get_default_broker()

    def get_default_slice(self) -> ISlice:
        """
        Get default inventory slice for the broker
        @return inventory slice for broker
        """
        slice_list = self.get_inventory_slices()
        if slice_list is not None and len(slice_list) > 0:
            return slice_list[0]
        return None

    def initialize(self):
        if not self.initialized:
            super().initialize()
            self.registry.set_slices_plugin(plugin=self.plugin)
            self.registry.initialize()
            self.initialized = True

    def issue_delayed(self):
        super().issue_delayed()

        self.extend_tickets_client(rset=self.extending)
        self.extending.clear()

        self.tickets_client(rset=self.ticketing)
        self.ticketing.clear()

    def ticket_client(self, *, reservation: IClientReservation):
        if not self.recovered:
            self.ticketing.add(reservation=reservation)
        else:
            self.wrapper.ticket(reservation=reservation, destination=self)

    def tickets_client(self, *, rset: ReservationSet):
        for reservation in rset.values():
            try:
                if isinstance(reservation, IBrokerReservation):
                    self.ticket_broker(reservation=reservation)
                elif isinstance(reservation, IClientReservation):
                    self.ticket_client(reservation=reservation)
                else:
                    self.logger.warning("Reservation #{} cannot be ticketed".format(reservation.get_reservation_id()))
            except Exception as e:
                self.logger.error("Could not ticket for #{} e: {}".format(reservation.get_reservation_id(), e))

    def ticket_broker(self, *, reservation: IBrokerReservation):
        if not self.recovered:
            self.ticketing.add(reservation=reservation)
        else:
            self.wrapper.ticket_request(reservation=reservation, caller=reservation.get_client_auth_token(),
                                        callback=reservation.get_callback(), compare_seq_numbers=False)

    def ticket(self, *, reservation: IReservation, callback: IClientCallbackProxy, caller: AuthToken):
        if not self.is_recovered() or self.is_stopped():
            raise BrokerException(Constants.invalid_actor_state)

        self.wrapper.ticket_request(reservation=reservation, caller=caller, callback=callback, compare_seq_numbers=True)

    def relinquish(self, *, reservation: IReservation, caller: AuthToken):
        if not self.is_recovered() or self.is_stopped():
            raise BrokerException(Constants.invalid_actor_state)
        self.wrapper.relinquish_request(reservation=reservation, caller=caller)

    def tick_handler(self):
        self.policy.allocate(cycle=self.current_cycle)
        self.bid(cycle=self.current_cycle)
        self.close_expiring(cycle=self.current_cycle)

    def update_ticket(self, *, reservation: IReservation, update_data, caller: AuthToken):
        if not self.is_recovered() or self.is_stopped():
            raise BrokerException(Constants.invalid_actor_state)
        self.wrapper.update_ticket(reservation=reservation, update_data=update_data, caller=caller)

    def update_delegation(self, *, delegation: IDelegation, update_data, caller: AuthToken):
        if not self.is_recovered() or self.is_stopped():
            raise BrokerException(Constants.invalid_actor_state)
        self.wrapper.update_delegation(delegation=delegation, update_data=update_data, caller=caller)

    def register_client(self, *, client: Client):
        database = self.plugin.get_database()

        try:
            database.get_client(guid=client.get_guid())
        except Exception as e:
            raise BrokerException(Constants.object_not_found.format(client.get_guid(), e))

        try:
            database.add_client(client=client)
        except Exception as e:
            raise BrokerException(Constants.failed_to_add.format(client, e))

    def unregister_client(self, *, guid: ID):
        database = self.plugin.get_database()

        try:
            database.get_client(guid=guid)
        except Exception as e:
            raise BrokerException(Constants.object_not_found.format(guid, e))

        try:
            database.remove_client(guid=guid)
        except Exception as e:
            raise BrokerException(Constants.failed_to_remove.format(guid, e))

    def get_client(self, *, guid: ID) -> Client:
        database = self.plugin.get_database()
        ret_val = None
        try:
            client_obj = database.get_client(guid=guid)
            ret_val = pickle.loads(client_obj.get(Constants.property_pickle_properties))
        except Exception as e:
            raise BrokerException(Constants.object_not_found.format(guid, e))

        return ret_val

    def modify(self, *, reservation_id: ID, modify_properties: dict):
        return

    @staticmethod
    def get_management_object_class() -> str:
        return BrokerManagementObject.__name__

    @staticmethod
    def get_management_object_module() -> str:
        return BrokerManagementObject.__module__

    @staticmethod
    def get_kafka_service_class() -> str:
        return BrokerService.__name__

    @staticmethod
    def get_kafka_service_module() -> str:
        return BrokerService.__module__

    @staticmethod
    def get_mgmt_kafka_service_class() -> str:
        return KafkaBrokerService.__name__

    @staticmethod
    def get_mgmt_kafka_service_module() -> str:
        return KafkaBrokerService.__module__



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod, ABC
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from fabric.actor.core.apis.i_actor import IActor
    from fabric.actor.core.apis.i_reservation import IReservation
    from fabric.actor.core.plugins.config.config_token import ConfigToken
    from fabric.actor.core.time.term import Term
    from fabric.actor.core.util.reservation_set import ReservationSet
    from fabric.actor.core.kernel.resource_set import ResourceSet
    from fabric.actor.core.util.id import ID
    from fabric.actor.core.apis.i_delegation import IDelegation


class IPolicy(ABC):
    """
    IPolicy encapsulates all policy decisions an actor must make to
    perform its functions. Each actor has one policy instance that controls its
    policy decisions. The policy instance must implement the IPolicy
    interface together with one or more of the other policy interfaces (depending
    on the role the actor performs).
    Some things to know when implementing a policy:
    - The policy is always called by a single thread. No synchronization is
    required, but method implementations should not block for long periods of
    time, as that would prevent the actor from processing other events.
    - The policy operates on the passed ResourceSets. ResourceSets encapsulate
    the free resource pools and resource requests. The policy satisfies requests
    by using ResourceSet methods to shift resources from one ResourceSet to
    another. Resource attributes etc. flow with these operations, and they drive
    other resource-specific operations including SHARP ticket delegation.
    - The policy has access to all ResourceData attributes attached to the
    request and the free resource pools. The client-side methods may decorate the
    request with attributes (e.g., bids), which are passed through to the
    server-side policy that handles the request. The client-side also has the
    option of touching the AuthToken passed with the request...this may be a bit
    goofy.
    - If it is necessary to defer an allocation request, return false. At some
    later time, when the policy is ready to satisfy the request (e.g., after an
    auction on an agent), set bidPending to false and arrange a call to
    probePending on the requesting reservation. This causes the reservation to
    retry the pending operation.
    """
    @abstractmethod
    def prepare(self, *, cycle: int):
        """
        Informs the policy that processing for a new cycle is about to begin. The
        policy should initialize whatever internal state is necessary to process
        a new cycle.
        Note: The cycle number parameter is redundant and is passed for
        convenience. The policy can always obtain the cycle number by calling
        IActor.getCurrentCycle()
        """

    @abstractmethod
    def finish(self, *, cycle: int):
        """
        Informs the policy that all processing for the specified cycle is
        complete. The policy can safely discard any state associated with the
        cycle or any previous cycles.
        Note: The cycle number parameter is redundant and is passed for
        convenience. The policy can always obtain the cycle number by calling
        IActor.getCurrentCycle()
        """

    @abstractmethod
    def extend(self, *, reservation: IReservation, resources: ResourceSet, term: Term):
        """
        Notifies the policy that a reservation is about to be extended. This
        method will be invoked only for reservations, whose extensions have not
        been triggered by the policy, e.g, from the management interface. The
        policy should update its state to reflect the extend request.

        @params reservation: reservation to be extended
        @params resources: resource set used for the extension
        @params term: term used for the extension
        """

    @abstractmethod
    def close(self, *, reservation: IReservation):
        """
        Notifies the policy that a reservation is about to be closed. This method
        will be invoked for every reservation that is about to be closed, even if
        the close was triggered by the policy itself. The policy should update
        its internal state/cancel pending operations associated with the
        reservation.

        @params reservation: reservation about to be closed
        """

    @abstractmethod
    def closed(self, *, reservation: IReservation):
        """
        Notifies the policy that a reservation has been closed. This method will
        be invoked for every reservation that closes successfully. The policy
        must uncommit any resources associated with the reservation, e.g,
        physical machines, currency, etc.
        Note: For an authority resources are released using the
        IAuthorityPolicy.release(ResourceSet) method. Authority policy
        implementations should not consider the resources of the passed
        reservation as released. The release will take place once all
        configuration actions complete.

        @params reservation: closed reservation
        """

    @abstractmethod
    def closed_delegation(self, *, delegation: IDelegation):
        """
        Notifies the policy that a delegation has been closed. This method will
        be invoked for every delegation that closes successfully. The policy
        must uncommit any resources associated with the delegation, e.g,
        physical machines, currency, etc.

        @params delegation: closed delegation
        """

    @abstractmethod
    def remove(self, *, reservation: IReservation):
        """
        Notifies the policy that a reservation is about to be removed. This
        method will be invoked for each reservation that is to be removed from
        the system. The policy should remove any state that it maintains for the
        reservation.
        Note: Only failed and closed reservations can be removed. The
        system will not invoke this method if the reservation is not closed or
        failed.

        @params reservation: reservation to remove
        """

    @abstractmethod
    def query(self, *, p) -> dict:
        """
        Answers a query from another actor. This method is intended to be used to
        obtain policy-specific parameters and information. This method should be
        used when writing more complex policies requiring additional interaction
        among actors. Instead of extending the proxies to support
        passing/obtaining the required information, policy code can use the query
        interface to request/obtain such information. The implementation should
        not block for prolonged periods of time. If necessary, future versions
        will update this interface to allow query responses to be delivered using
        callbacks.

        @params p : a properties list of query parameters. Can be null or empty.

        @returns a properties list of outgoing values. If the incoming properties
                collection is null or empty, should return all possible
                properties that can be relevant to the caller.
        """

    @abstractmethod
    def configuration_complete(self, *, action: str, token: ConfigToken, out_properties: dict):
        """
        Notifies the policy that a configuration action for the object
        represented by the token parameter has completed.

        @params action : configuration action. See Config.Target*
        @params token : object or a token for the object whose configuration action has completed
        @params out_properties : output properties produced by the configuration action
        """

    @abstractmethod
    def reset(self):
        """
        Post recovery entry point. This method will be invoked once all revisit
        operations are complete and the actor is ready to operate normally.
        @raises Exception in case of error
        """

    @abstractmethod
    def recovery_starting(self):
        """
        Informs the policy that recovery is about to begin.
        """

    @abstractmethod
    def revisit(self, *, reservation: IReservation):
        """
        Informs the policy about a reservation. Called during recovery/startup.
        The policy must re-establish any state required for the management of the
        reservation.

        @params reservation: reservation being recovered
        @raises Exception in case of error
        """

    @abstractmethod
    def revisit_delegation(self, *, delegation: IDelegation):
        """
        Informs the policy about a delegation. Called during recovery/startup.
        The policy must re-establish any state required for the management of the
        delegation.

        @params delegation: delegation being recovered
        @raises Exception in case of error
        """

    @abstractmethod
    def recovery_ended(self):
        """
        Informs the policy that recovery has completed.
        """

    @abstractmethod
    def set_actor(self, *, actor: IActor):
        """
        Sets the actor the policy belongs to.

        @params actor : the actor the policy belongs to
        """

    @abstractmethod
    def get_closing(self, *, cycle: int) -> ReservationSet:
        """
        Returns a set of reservations that must be closed.

        @params cycle: the current cycle

        @returns reservations to be closed
        """

    @abstractmethod
    def get_guid(self) -> ID:
        """
        Returns the globally unique identifier of this policy object instance.

        @returns guid
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import unittest

from fabric.actor.core.policy.free_allocated_set import FreeAllocatedSet


class FreeAllocatedSetTest(unittest.TestCase):
    def test_it(self):
        free_set = FreeAllocatedSet()
        self.assertEqual(0, len(free_set.free_set))
        self.assertEqual(0, len(free_set.allocated))

        for i in range(50):
            self.assertEqual(i, free_set.get_free())
            self.assertEqual(0, free_set.get_allocated())

            self.assertFalse(free_set.free_set.__contains__(i))
            self.assertFalse(free_set.allocated.__contains__(i))
            free_set.add_inventory(item=i)
            self.assertTrue(free_set.free_set.__contains__(i))
            self.assertFalse(free_set.allocated.__contains__(i))

        i = free_set.allocate()
        self.assertIsNotNone(i)
        self.assertEqual(1, free_set.get_allocated())
        self.assertEqual(49, free_set.get_free())
        self.assertTrue(free_set.allocated.__contains__(i))
        self.assertFalse(free_set.free_set.__contains__(i))

        free_set.free()
        self.assertEqual(0, free_set.get_allocated())
        self.assertEqual(50, free_set.get_free())
        self.assertFalse(free_set.allocated.__contains__(i))
        self.assertTrue(free_set.free_set.__contains__(i))

        ll = free_set.allocate_count(count=25)
        self.assertEqual(25, len(ll))

        for i in ll:
            self.assertEqual(25, free_set.get_allocated())
            self.assertEqual(25, free_set.get_free())
            self.assertTrue(free_set.allocated.__contains__(i))
            self.assertFalse(free_set.free_set.__contains__(i))

        sub1 = ll[0:10]
        sub2 = ll[10:25]
        free_set.free_list(items=sub1)

        self.assertEqual(35, free_set.get_free())
        self.assertEqual(15, free_set.get_allocated())

        for i in sub1:
            self.assertTrue(free_set.free_set.__contains__(i))
            self.assertFalse(free_set.allocated.__contains__(i))

        for i in sub2:
            self.assertFalse(free_set.free_set.__contains__(i))
            self.assertTrue(free_set.allocated.__contains__(i))

        free_set.free_list(items=sub2)

        self.assertEqual(50, free_set.get_free())
        self.assertEqual(0, free_set.get_allocated())

        for i in ll:
            self.assertTrue(free_set.free_set.__contains__(i))
            self.assertFalse(free_set.allocated.__contains__(i))




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import traceback
from typing import List

from fabric.actor.core.apis.i_mgmt_authority import IMgmtAuthority
from fabric.actor.core.apis.i_reservation import ReservationCategory
from fabric.actor.core.common.constants import Constants, ErrorCodes
from fabric.actor.core.manage.kafka.kafka_server_actor import KafkaServerActor
from fabric.actor.core.util.id import ID
from fabric.message_bus.messages.get_reservations_request_avro import GetReservationsRequestAvro
from fabric.message_bus.messages.get_reservation_units_request_avro import GetReservationUnitsRequestAvro
from fabric.message_bus.messages.get_unit_request_avro import GetUnitRequestAvro
from fabric.message_bus.messages.reservation_mng import ReservationMng
from fabric.message_bus.messages.result_avro import ResultAvro
from fabric.message_bus.messages.unit_avro import UnitAvro


class KafkaAuthority (KafkaServerActor, IMgmtAuthority):
    def get_authority_reservations(self, *, id_token: str = None) -> List[ReservationMng]:
        self.clear_last()
        status = ResultAvro()
        rret_val = None

        try:
            request = GetReservationsRequestAvro()
            request.guid = str(self.management_id)
            request.auth = self.auth
            request.message_id = str(ID())
            request.callback_topic = self.callback_topic
            request.reservation_type = ReservationCategory.Authority.name
            request.id_token = id_token

            ret_val = self.producer.produce_sync(topic=self.kafka_topic, record=request)

            self.logger.debug(Constants.management_inter_actor_outbound_message.format(request.name, self.kafka_topic))

            if ret_val:
                message_wrapper = self.message_processor.add_message(message=request)

                with message_wrapper.condition:
                    message_wrapper.condition.wait(Constants.management_api_timeout_in_seconds)

                if not message_wrapper.done:
                    self.logger.debug(Constants.management_api_timeout_occurred)
                    self.message_processor.remove_message(msg_id=request.get_message_id())
                    status.code = ErrorCodes.ErrorTransportTimeout.value
                    status.message = ErrorCodes.ErrorTransportTimeout.name
                else:
                    self.logger.debug(Constants.management_inter_actor_inbound_message.format(message_wrapper.response))
                    status = message_wrapper.response.status
                    if status.code == 0:
                        rret_val = message_wrapper.response.reservations
            else:
                self.logger.debug(Constants.management_inter_actor_message_failed.format(
                    request.name, self.kafka_topic))
                status.code = ErrorCodes.ErrorTransportFailure.value
                status.message = ErrorCodes.ErrorTransportFailure.name

        except Exception as e:
            self.last_exception = e
            status.code = ErrorCodes.ErrorInternalError.value
            status.message = ErrorCodes.ErrorInternalError.name
            status.details = traceback.format_exc()

        self.last_status = status

        return rret_val

    def get_reservation_units(self, *, rid: ID, id_token: str = None) -> List[UnitAvro]:
        self.clear_last()
        status = ResultAvro()
        rret_val = None

        try:
            request = GetReservationUnitsRequestAvro()
            request.guid = str(self.management_id)
            request.auth = self.auth
            request.message_id = str(ID())
            request.callback_topic = self.callback_topic
            request.reservation_id = str(rid)
            request.id_token = id_token

            ret_val = self.producer.produce_sync(topic=self.kafka_topic, record=request)

            self.logger.debug(Constants.management_inter_actor_outbound_message.format(request.name, self.kafka_topic))

            if ret_val:
                message_wrapper = self.message_processor.add_message(message=request)

                with message_wrapper.condition:
                    message_wrapper.condition.wait(Constants.management_api_timeout_in_seconds)

                if not message_wrapper.done:
                    self.logger.debug(Constants.management_api_timeout_occurred)
                    self.message_processor.remove_message(msg_id=request.get_message_id())
                    status.code = ErrorCodes.ErrorTransportTimeout.value
                    status.message = ErrorCodes.ErrorTransportTimeout.name
                else:
                    self.logger.debug(Constants.management_inter_actor_inbound_message.format(message_wrapper.response))
                    status = message_wrapper.response.status
                    if status.code == 0:
                        rret_val = message_wrapper.response.units
            else:
                self.logger.debug(Constants.management_inter_actor_message_failed.format(
                    request.name, self.kafka_topic))
                status.code = ErrorCodes.ErrorTransportFailure.value
                status.message = ErrorCodes.ErrorTransportFailure.name

        except Exception as e:
            self.last_exception = e
            status.code = ErrorCodes.ErrorInternalError.value
            status.message = ErrorCodes.ErrorInternalError.name
            status.details = traceback.format_exc()

        self.last_status = status

        return rret_val

    def get_reservation_unit(self, *, uid: ID, id_token: str = None) -> UnitAvro:
        self.clear_last()
        status = ResultAvro()
        rret_val = None

        try:
            request = GetUnitRequestAvro()
            request.guid = str(self.management_id)
            request.auth = self.auth
            request.message_id = str(ID())
            request.callback_topic = self.callback_topic
            request.reservation_id = str(uid)
            request.id_token = id_token

            ret_val = self.producer.produce_sync(topic=self.kafka_topic, record=request)

            self.logger.debug(Constants.management_inter_actor_outbound_message.format(
                request.name, self.kafka_topic))

            if ret_val:
                message_wrapper = self.message_processor.add_message(message=request)

                with message_wrapper.condition:
                    message_wrapper.condition.wait(Constants.management_api_timeout_in_seconds)

                if not message_wrapper.done:
                    self.logger.debug(Constants.management_api_timeout_occurred)
                    self.message_processor.remove_message(msg_id=request.get_message_id())
                    status.code = ErrorCodes.ErrorTransportTimeout.value
                    status.message = ErrorCodes.ErrorTransportTimeout.name
                else:
                    self.logger.debug(Constants.management_inter_actor_inbound_message.format(message_wrapper.response))
                    status = message_wrapper.response.status
                    if status.code == 0:
                        rret_val = message_wrapper.response.units
            else:
                self.logger.debug(Constants.management_inter_actor_message_failed.format(
                    request.name, self.kafka_topic))
                status.code = ErrorCodes.ErrorTransportFailure.value
                status.message = ErrorCodes.ErrorTransportFailure.name

        except Exception as e:
            self.last_exception = e
            status.code = ErrorCodes.ErrorInternalError.value
            status.message = ErrorCodes.ErrorInternalError.name
            status.details = traceback.format_exc()

        self.last_status = status

        if rret_val is not None:
            return rret_val.__iter__().__next__()

        return rret_val

    def clone(self):
        return KafkaAuthority(guid=self.management_id,
                              kafka_topic=self.kafka_topic,
                              auth=self.auth, logger=self.logger,
                              message_processor=self.message_processor,
                              producer=self.producer)


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import pickle
import threading

from fabric.actor.core.plugins.db.actor_database import ActorDatabase
from fabric.actor.core.plugins.db.client_database import ClientDatabase
from fabric.actor.core.util.client import Client
from fabric.actor.core.util.id import ID


class ServerActorDatabase(ActorDatabase, ClientDatabase):
    def add_client(self, *, client: Client):
        try:
            self.lock.acquire()
            properties = pickle.dumps(client)
            self.db.add_client(act_id=self.actor_id, clt_name=client.get_name(), clt_guid=str(client.get_guid()),
                               properties=properties)
        finally:
            self.lock.release()

    def update_client(self, *, client: Client):
        try:
            self.lock.acquire()
            properties = pickle.dumps(client)
            self.db.update_client(act_id=self.actor_id, clt_name=client.get_name(), properties=properties)
        finally:
            self.lock.release()

    def remove_client(self, *, guid: ID):
        try:
            self.lock.acquire()
            self.db.remove_client_by_guid(act_id=self.actor_id, clt_guid=str(guid))
        finally:
            self.lock.release()

    def get_client(self, *, guid: ID) -> dict:
        result = None
        try:
            self.lock.acquire()
            result = self.db.get_client_by_guid(act_id=self.actor_id, clt_guid=str(guid))
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return result

    def get_clients(self) -> list:
        result = None
        try:
            self.lock.acquire()
            result = self.db.get_clients(act_id=self.actor_id)
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return result



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.common.exceptions import PolicyException


class FreeAllocatedSet:
    """
    FreeAllocatedSet is a simple data structure that maintains two sets: free and allocated.
    The structure can be used to track free and allocated items. Use add_inventory to
    add inventory items, allocate to allocate an item, and free to free an item.
    """
    def __init__(self):
        # Free set.
        self.free_set = set()
        # Allocated set.
        self.allocated = set()

    def add_inventory(self, *, item):
        """
        Adds an inventory item to the set.
        @param item item to add
        """
        if item is None:
            raise PolicyException("item cannot be null")

        if item in self.allocated:
            raise PolicyException("item is already in allocated")

        self.free_set.add(item)

    def allocate(self, *, tag=None, config_tag: bool = None, count: int = None):
        """
        Allocates an item from the set.
        @param tag tag
        @param config_tag config_tag
        @param count number of units to allocate.

        @return an allocated item, or null if the set does not have a free item
        """
        item = None

        if len(self.free_set) > 0:
            if tag is None and config_tag is None and count is None:
                item = self.free_set.pop()
                self.allocated.add(item)

            elif tag is not None and config_tag is not None:
                if config_tag:
                    item = tag
                    if item not in self.free_set:
                        raise PolicyException("item is already in allocated: {}".format(item))
                else:
                    item = self.free_set.pop()
                self.allocated.add(item)
        return item

    def allocate_count(self, *, count: int):
        item = None
        if len(self.free_set) > 0:
            item = []
            for i in range(count):
                val = self.free_set.pop()
                self.allocated.add(val)
                item.append(val)
        return item

    def free(self, *, item=None, count: int = None):
        if item is None and count is None:
            if len(self.allocated) == 0:
                raise PolicyException("no items have been allocated")

            item = self.allocated.pop()
            self.free_set.add(item)
        elif item is not None:
            if item not in self.allocated:
                raise PolicyException("item has not been allocated")
            if item in self.free_set:
                raise PolicyException("item has already been freed")

            self.allocated.remove(item)
            self.free_set.add(item)
        elif count is not None:
            for i in range(count):
                if len(self.allocated) == 0:
                    raise PolicyException("no items have been allocated")

                item = self.allocated.pop()
                self.free_set.add(item)

    def free_list(self, *, items: list):
        for i in items:
            self.free(item=i)

    def get_free(self) -> int:
        return len(self.free_set)

    def get_allocated(self) -> int:
        return len(self.allocated)

    def size(self):
        return self.get_free() + self.get_allocated()

    def __str__(self):
        return "Free: {}\nAllocated: {}".format(self.free_set, self.allocated)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.apis.i_concrete_set import IConcreteSet
from fabric.actor.core.core.controller import Controller
from fabric.actor.core.core.ticket import Ticket
from fabric.actor.core.kernel.reservation_client import ReservationClient
from fabric.actor.core.kernel.reservation_states import ReservationStates, ReservationPendingStates
from fabric.actor.core.time.term import Term
from fabric.actor.core.util.resource_type import ResourceType


class ControllerTestWrapper(Controller):
    def bid(self):
        candidates = self.policy.formulate_bids(cycle=self.current_cycle)
        if candidates is not None:
            ticketing = candidates.get_ticketing()
            if ticketing is not None:
                for r in ticketing.values():
                    print("cycle: {} Ticket request for: {}".format(self.current_cycle, r))

                    if r.get_slice_name().startswith("fail"):
                        already_failed = False
                        slice_obj = r.get_slice()
                        for slice_reservation in slice_obj.get_reservations().values():
                            if slice_reservation.get_state() == ReservationStates.Failed:
                                already_failed = True
                                break
                        if not already_failed:
                            self.fail_ticket(r)
                            continue

                    delegation = self.get_plugin().get_ticket_factory().make_delegation(
                        units=r.get_approved_resources().get_units(),
                        term=r.get_approved_term(),
                        rtype=r.get_approved_type(),
                        vector= None)
                    ticket = self.get_plugin().get_ticket_factory().make_ticket(delegation=delegation, source=None)
                    cs = Ticket(ticket=ticket, plugin=self.get_plugin())
                    self.update_ticket_wrapper(r, r.get_approved_type(), r.get_approved_units(), cs,
                                               r.get_approved_term())

            extending = candidates.get_extending()
            if extending is not None:
                for r in extending.values():
                    print("cycle: {} Extend Ticket request for: {}".format(self.current_cycle, r))
                    delegation = self.get_plugin().get_ticket_factory().make_delegation(
                        units=r.get_approved_resources().get_units(),
                        term=r.get_approved_term(),
                        rtype=r.get_approved_type(), vector=None)
                    ticket = self.get_plugin().get_ticket_factory().make_ticket(delegation=delegation, source=None)
                    cs = Ticket(ticket=ticket, plugin=self.get_plugin())
                    self.update_ticket_wrapper(r, r.get_approved_type(), r.get_approved_units(), cs,
                                               r.get_approved_term())

    def close_expiring(self):
        rset = self.policy.get_closing(cycle=self.current_cycle)
        if rset is not None:
            for r in rset.values():
                print("cycle: {} closing reservation r: {}".format(self.current_cycle, r))
                r.transition(prefix="close", state=ReservationStates.Closed, pending=ReservationPendingStates.None_)

    def process_redeeming(self):
        rset = self.policy.get_redeeming(cycle=self.current_cycle)
        if rset is not None:
            for r in rset.values():
                if r.get_state() == ReservationStates.Ticketed:
                    print("cycle: {} redeeming reservation r: {}".format(self.current_cycle, r))
                else:
                    print("cycle: {} extending lease for reservation r: {}".format(self.current_cycle, r))

                delegation = self.get_plugin().get_ticket_factory().make_delegation(units=r.resources.get_units(),
                                                                                    term=r.term,
                                                                                    rtype=r.resources.get_type(), vector=None)
                ticket = self.get_plugin().get_ticket_factory().make_ticket(delegation=delegation, source=None)
                cs = Ticket(ticket=ticket, plugin=self.get_plugin())
                self.update_lease_wrapper(r, r.get_approved_type(), r.get_approved_units(), cs, r.get_approved_term())

    def update_lease_wrapper(self, reservation: ReservationClient, rtype: ResourceType, units: int, cs: IConcreteSet, term: Term):
        if reservation.state == ReservationStates.Ticketed:
            reservation.leased_resources = reservation.resources.abstract_clone()
            reservation.leased_resources.units = units
            reservation.leased_resources.type = rtype
            reservation.leased_resources.set_resources(cset=cs)

            reservation.previous_lease_term = None
            reservation.previous_term = reservation.term
            reservation.lease_term = term.clone()
            reservation.term = reservation.lease_term

            reservation.transition(prefix="redeem", state=ReservationStates.Active, pending=ReservationPendingStates.None_)
        else:
            reservation.leased_resources.units = units
            reservation.leased_resources.type = rtype
            reservation.leased_resources.resources.change(concrete_set=cs, configure=False)

            reservation.previous_lease_term = reservation.requested_term
            reservation.previous_term = reservation.term
            reservation.lease_term = term.clone()
            reservation.term = reservation.lease_term

            reservation.transition(prefix="redeem", state=ReservationStates.Active, pending=ReservationPendingStates.None_)

    def update_ticket_wrapper(self, reservation: ReservationClient, rtype: ResourceType, units: int,
                              ticket: Ticket, term: Term):
        if reservation.state == ReservationStates.Nascent:
            reservation.resources = reservation.get_approved_resources().abstract_clone()
            reservation.resources.units = units
            reservation.resources.type = rtype
            reservation.resources.set_resources(cset=ticket)

            reservation.previous_ticket_term = None
            reservation.previous_term = None
            reservation.term = term.clone()
            reservation.ticket_term = reservation.term

            reservation.transition(prefix="ticket", state=ReservationStates.Ticketed, pending=ReservationPendingStates.None_)
        else:
            reservation.resources.units = units
            reservation.resources.type = rtype
            reservation.resources.resources.change(concrete_set=ticket, configure=False)

            reservation.previous_term = reservation.term
            reservation.previous_ticket_term = reservation.ticket_term
            reservation.term = term.clone()
            reservation.ticket_term = reservation.term

            reservation.transition(prefix="extendticket", state=ReservationStates.ActiveTicketed, pending=ReservationPendingStates.None_)

    def fail_ticket(self, r: ReservationClient):
        r.transition(prefix="fail", state=ReservationStates.Failed, pending=ReservationPendingStates.None_)


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from typing import TYPE_CHECKING

from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import PluginException
from fabric.actor.core.plugins.config.config import Config
from fabric.actor.core.apis.i_substrate_database import ISubstrateDatabase
from fabric.actor.core.plugins.base_plugin import BasePlugin
from fabric.actor.core.apis.i_substrate import ISubstrate
from fabric.actor.core.util.prop_list import PropList

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_database import IDatabase
    from fabric.actor.core.apis.i_reservation import IReservation
    from fabric.actor.core.core.unit import Unit
    from fabric.actor.core.plugins.config.config_token import ConfigToken


class Substrate(BasePlugin, ISubstrate):
    def initialize(self):
        super().initialize()
        if not isinstance(self.db, ISubstrateDatabase):
            raise PluginException("Substrate database class must implement ISubstrateDatabase")

    def transfer_in(self, *, reservation: IReservation, unit: Unit):
        try:
            # record the node in the database
            self.get_substrate_database().add_unit(u=unit)
            # prepare the transfer
            self.prepare_transfer_in(reservation=reservation, unit=unit)
            # update the unit database record
            # since prepareTransferIn may have added new properties.
            self.db.update_unit(u=unit)
            # perform the node configuration
            self.do_transfer_in(reservation=reservation, unit=unit)
        except Exception as e:
            self.fail_and_update(unit=unit, message="transferIn error", e=e)

    def transfer_out(self, *, reservation: IReservation, unit: Unit):
        try:
            # prepare the transfer out
            self.prepare_transfer_out(reservation=reservation, unit=unit)
            # update the unit database record
            self.get_substrate_database().update_unit(u=unit)
            # perform the node configuration
            self.do_transfer_out(reservation=reservation, unit=unit)
        except Exception as e:
            self.fail_and_update(unit=unit, message="transferOut error", e=e)

    def modify(self, *, reservation: IReservation, unit: Unit):
        try:
            # prepare the transfer out
            self.prepare_modify(reservation=reservation, unit=unit)
            # update the unit database record
            self.db.update_unit(u=unit)
            # perform the node configuration
            self.do_modify(reservation=reservation, unit=unit)
        except Exception as e:
            self.fail_and_update(unit=unit, message="modify error", e=e)

    def prepare_transfer_in(self, *, reservation: IReservation, unit: Unit):
        """
        Performs additional setup operations before configuring the unit. This is
        an optional step of the transfer in process and can be used to set custom
        properties on the reservation/unit objects before invoking the
        configuration subsystem. For example, if the policy did not allocate an
        IP address for the unit, but an IP address is required to configure the
        unit, the IP address can be allocated in this function.
        @param reservation reservation containing the unit
        @param unit unit to prepare
        @throws Exception in case of error
        """

    def prepare_transfer_out(self, *, reservation: IReservation, unit: Unit):
        """
        Prepares the unit for transfer out. Note: resources assigned in
        prepareTransferIn cannot be released yet, since they are still in use.
        The resources can be released only when the transferOut operation
        completes. See {@link #processLeaveComplete(Object, Properties)}.
        @param reservation reservation containing the unit
        @param unit unit to prepare
        @throws Exception in case of error
        """

    def prepare_modify(self, *, reservation: IReservation, unit: Unit):
        """
        Prepares the unit for modification.
        @param reservation reservation containing the unit
        @param unit unit to prepare
        @throws Exception in case of error
        """

    def get_config_properties_from_reservation(self, *, reservation: IReservation, unit: Unit) -> dict:
        temp = reservation.get_resources().get_local_properties()
        temp = PropList.merge_properties(incoming=reservation.get_slice().get_local_properties(), outgoing=temp)

        if self.is_site_authority():
            temp = PropList.merge_properties(incoming=reservation.get_resources().get_config_properties(),
                                             outgoing=temp)
            temp = PropList.merge_properties(incoming=reservation.get_slice().get_config_properties(),
                                             outgoing=temp)

            if reservation.get_requested_resources() is not None:
                ticket = reservation.get_requested_resources().get_resources()
                temp = PropList.merge_properties(incoming=ticket.get_properties(), outgoing=temp)
                rticket = ticket.get_ticket()
                temp = PropList.merge_properties(incoming=rticket.get_properties(), outgoing=temp)

        temp = PropList.merge_properties(incoming=unit.get_properties(), outgoing=temp)

        return temp

    def do_transfer_in(self, *, reservation: IReservation, unit: Unit):
        prop = self.get_config_properties_from_reservation(reservation=reservation, unit=unit)
        self.config.create(token=unit, properties=prop)

    def do_transfer_out(self, *, reservation: IReservation, unit: Unit):
        prop = self.get_config_properties_from_reservation(reservation=reservation, unit=unit)
        self.config.delete(token=unit, properties=prop)

    def do_modify(self, *, reservation: IReservation, unit: Unit):
        prop = self.get_config_properties_from_reservation(reservation=reservation, unit=unit)
        self.config.modify(token=unit, properties=prop)

    def fail_and_update(self, *, unit: Unit, message: str, e: Exception):
        self.logger.error(message)
        self.logger.error(str(e))

        try:
            unit.fail(message=message, exception=e)
            self.db.update_unit(u=unit)
        except Exception as e:
            self.logger.error("could not update unit in database")
            self.logger.error(e)

    def fail_no_update(self, *, unit: Unit, message: str, e: Exception = None):
        self.logger.error(message)
        if e:
            self.logger.error(e)
        unit.fail(message=message, exception=e)

    def fail_modify_no_update(self, *, unit: Unit, message: str, e: Exception = None):
        self.logger.error(message)
        if e:
            self.logger.error(e)
        unit.fail_on_modify(message=message, exception=e)

    def merge_unit_properties(self, *, unit: Unit, properties: dict):
        # TODO
        return

    def process_create_complete(self, *, token: ConfigToken, properties: dict):
        self.logger.debug("Join")
        self.logger.debug(properties)

        if self.actor.is_stopped():
            raise PluginException(Constants.invalid_actor_state)

        sequence = Config.get_action_sequence_number(properties=properties)
        notice = None
        # TODO synchronized on token
        if sequence != token.get_sequence():
            self.logger.warning("(create complete) sequences mismatch: incoming ({}) local: ({}). Ignoring event.".
                                format(sequence, token.get_sequence()))
            return
        else:
            self.logger.debug("(create complete) incoming ({}) local: ({})".format(sequence, token.get_sequence()))

        result = Config.get_result_code(properties=properties)
        msg = Config.get_exception_message(properties=properties)
        if msg is None:
            msg = Config.get_result_code_message(properties=properties)

        if result == 0:
            self.logger.debug("create code 0 (success)")
            self.merge_unit_properties(unit=token, properties=properties)
            token.activate()

        elif result == -1:
            self.logger.debug("create code -1 with message: {}".format(msg))
            notice = "Exception during create for unit: {} {}".format(token.get_id(), msg)
            self.fail_no_update(unit=token, message=notice)

        else:
            self.logger.debug("create code {} with message: {}".format(result, msg))
            notice = "Error code= {} during create for unit: {} with message: {}".format(result, token.get_id(), msg)
            self.fail_no_update(unit=token, message=notice)

        try:
            self.get_substrate_database().update_unit(u=token)
        except Exception as e:
            self.logger.error(e)
        finally:
            self.logger.debug("process create complete")

    def process_delete_complete(self, *, token: ConfigToken, properties: dict):
        self.logger.debug("Leave")
        self.logger.debug(properties)

        if self.actor.is_stopped():
            raise PluginException(Constants.invalid_actor_state)

        sequence = Config.get_action_sequence_number(properties=properties)
        notice = None
        # TODO synchronized on token
        if sequence != token.get_sequence():
            self.logger.warning("(delete complete) sequences mismatch: incoming ({}) local: ({}). "
                                "Ignoring event.".format(sequence, token.get_sequence()))
            return
        else:
            self.logger.debug("(delete complete) incoming ({}) local: ({})".format(sequence, token.get_sequence()))

        result = Config.get_result_code(properties=properties)
        msg = Config.get_exception_message(properties=properties)
        if msg is None:
            msg = Config.get_result_code_message(properties=properties)

        if result == 0:
            self.logger.debug("delete code 0 (success)")
            self.merge_unit_properties(unit=token, properties=properties)
            token.close()

        elif result == -1:
            self.logger.debug("delete code -1 with message: {}".format(msg))
            notice = "Exception during create for unit: {} {}".format(token.get_id(), msg)
            self.fail_no_update(unit=token, message=notice)

        else:
            self.logger.debug("delete code {} with message: {}".format(result, msg))
            notice = "Error code= {} during create for unit: {} with message: {}".format(result, token.get_id(), msg)
            self.fail_no_update(unit=token, message=notice)

        try:
            self.get_substrate_database().update_unit(u=token)
        except Exception as e:
            self.logger.error(e)
        finally:
            self.logger.debug("process delete complete")

    def process_modify_complete(self, *, token: ConfigToken, properties: dict):
        self.logger.debug("Modify")
        self.logger.debug(properties)

        if self.actor.is_stopped():
            raise PluginException(Constants.invalid_actor_state)

        sequence = Config.get_action_sequence_number(properties=properties)
        notice = None
        # TODO synchronized on token
        if sequence != token.get_sequence():
            self.logger.warning("(modify complete) sequences mismatch: incoming ({}) local: ({}). "
                                "Ignoring event.".format(sequence, token.get_sequence()))
            return
        else:
            self.logger.debug("(modify complete) incoming ({}) local: ({})".format(sequence, token.get_sequence()))

        # TODO properties

        result = Config.get_result_code(properties=properties)
        msg = Config.get_exception_message(properties=properties)
        if msg is None:
            msg = Config.get_result_code_message(properties=properties)

        if result == 0:
            self.logger.debug("modify code 0 (success)")
            self.merge_unit_properties(unit=token, properties=properties)
            token.complete_modify()

        elif result == -1:
            self.logger.debug("modify code -1 with message: {}".format(msg))
            notice = "Exception during modify for unit: {} {}".format(token.get_id(), msg)
            self.fail_modify_no_update(unit=token, message=notice)

        else:
            self.logger.debug("modify code {} with message: {}".format(result, msg))
            notice = "Error code= {} during modify for unit: {} with message: {}".format(result, token.get_id(), msg)
            self.fail_modify_no_update(unit=token, message=notice)

        try:
            self.get_substrate_database().update_unit(u=token)
        except Exception as e:
            self.logger.error(e)
        finally:
            self.logger.debug("process modify complete")

    def get_substrate_database(self) -> ISubstrateDatabase:
        return self.db

    def set_database(self, *, db: IDatabase):
        if db is not None and not isinstance(db, ISubstrateDatabase):
            raise PluginException("db must implement ISubstrateDatabase")

        super().set_database(db=db)

    def update_props(self, *, reservation: IReservation, unit: Unit):
        try:
            self.get_substrate_database().update_unit(u=unit)
        except Exception as e:
            self.fail_and_update(unit=unit, message="update properties error", e=e)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING

from fabric.actor.core.apis.i_authority_proxy import IAuthorityProxy
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import ProxyException
from fabric.actor.core.kernel.rpc_request_type import RPCRequestType
from fabric.actor.core.proxies.kafka.kafka_broker_proxy import KafkaBrokerProxy
from fabric.actor.core.proxies.kafka.kafka_proxy import KafkaProxyRequestState
from fabric.actor.core.proxies.kafka.translate import Translate
from fabric.message_bus.messages.close_avro import CloseAvro
from fabric.message_bus.messages.extend_lease_avro import ExtendLeaseAvro
from fabric.message_bus.messages.modify_lease_avro import ModifyLeaseAvro
from fabric.message_bus.messages.redeem_avro import RedeemAvro
from fabric.message_bus.messages.reservation_avro import ReservationAvro

if TYPE_CHECKING:
    from fabric.actor.security.auth_token import AuthToken
    from fabric.actor.core.apis.i_rpc_request_state import IRPCRequestState
    from fabric.actor.core.apis.i_controller_callback_proxy import IControllerCallbackProxy
    from fabric.actor.core.apis.i_controller_reservation import IControllerReservation
    from fabric.actor.core.apis.i_reservation import IReservation


class KafkaAuthorityProxy(KafkaBrokerProxy, IAuthorityProxy):
    def __init__(self, *, kafka_topic: str, identity: AuthToken, logger):
        super().__init__(kafka_topic=kafka_topic, identity=identity, logger=logger)
        self.type = self.TypeSite

    def execute(self, *, request: IRPCRequestState):
        avro_message = None
        if request.get_type() == RPCRequestType.Redeem:
            avro_message = RedeemAvro()
            avro_message.message_id = str(request.get_message_id())
            avro_message.callback_topic = request.callback_topic
            avro_message.reservation = request.reservation
            avro_message.auth = Translate.translate_auth_to_avro(auth=request.caller)

        elif request.get_type() == RPCRequestType.ExtendLease:
            avro_message = ExtendLeaseAvro()
            avro_message.message_id = str(request.get_message_id())
            avro_message.callback_topic = request.callback_topic
            avro_message.reservation = request.reservation
            avro_message.auth = Translate.translate_auth_to_avro(auth=request.caller)

        elif request.get_type() == RPCRequestType.ModifyLease:
            avro_message = ModifyLeaseAvro()
            avro_message.message_id = str(request.get_message_id())
            avro_message.callback_topic = request.callback_topic
            avro_message.reservation = request.reservation
            avro_message.auth = Translate.translate_auth_to_avro(auth=request.caller)

        elif request.get_type() == RPCRequestType.Close:
            avro_message = CloseAvro()
            avro_message.message_id = str(request.get_message_id())
            avro_message.callback_topic = request.callback_topic
            avro_message.reservation = request.reservation
            avro_message.auth = Translate.translate_auth_to_avro(auth=request.caller)

        else:
            super().execute(request=request)
            return

        if self.producer is None:
            self.producer = self.create_kafka_producer()

        if self.producer is not None and self.producer.produce_sync(topic=self.kafka_topic, record=avro_message):
            self.logger.debug("Message {} written to {}".format(avro_message.name, self.kafka_topic))
        else:
            self.logger.error("Failed to send message {} to {} via producer {}".format(avro_message.name,
                                                                                       self.kafka_topic, self.producer))

    def _prepare(self, *, reservation: IControllerReservation, callback: IControllerCallbackProxy,
                 caller: AuthToken) -> IRPCRequestState:
        request = KafkaProxyRequestState()
        request.callback_topic = callback.get_kafka_topic()
        request.reservation = self.pass_authority_reservation(reservation=reservation, caller=caller)
        request.caller = caller
        return request

    def prepare_redeem(self, *, reservation: IControllerReservation, callback: IControllerCallbackProxy,
                       caller: AuthToken) -> IRPCRequestState:
        return self._prepare(reservation=reservation, callback=callback, caller=caller)

    def prepare_extend_lease(self, *, reservation: IControllerReservation, callback: IControllerCallbackProxy,
                             caller: AuthToken) -> IRPCRequestState:
        return self._prepare(reservation=reservation, callback=callback, caller=caller)

    def prepare_modify_lease(self, *, reservation: IControllerReservation, callback: IControllerCallbackProxy,
                             caller: AuthToken) -> IRPCRequestState:
        return self._prepare(reservation=reservation, callback=callback, caller=caller)

    def prepare_close(self, *, reservation: IControllerReservation, callback: IControllerCallbackProxy,
                      caller: AuthToken) -> IRPCRequestState:
        return self._prepare(reservation=reservation, callback=callback, caller=caller)

    @staticmethod
    def pass_authority_reservation(reservation: IReservation, caller: AuthToken) -> ReservationAvro:
        cs = reservation.get_resources().get_resources()
        if cs is None:
            raise ProxyException(Constants.not_specified_prefix.format("ticket"))

        avro_reservation = ReservationAvro()
        avro_reservation.slice = Translate.translate_slice_to_avro(slice_obj=reservation.get_slice())
        avro_reservation.term = Translate.translate_term(term=reservation.get_requested_term())
        avro_reservation.reservation_id = str(reservation.get_reservation_id())
        avro_reservation.sequence = reservation.get_lease_sequence_out()

        rset = Translate.translate_resource_set(resource_set=reservation.get_resources(),
                                                direction=Translate.direction_authority)
        cset = reservation.get_requested_resources().get_resources()

        encoded = None
        if cset is not None:
            encoded = cset.encode(protocol=Constants.protocol_kafka)
            if encoded is None:
                raise ProxyException("Unsupported IConcreteSet: {}".format(type(cset)))

        rset.concrete = encoded

        avro_reservation.resource_set = rset

        return avro_reservation



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import json
import os
from enum import Enum
from typing import List

import re
import requests

from fabric.actor.core.apis.i_actor import ActorType


class PdpAuthException(Exception):
    """
    PDP Auth Exception
    """


class ResourceType(Enum):
    """
    Resource Type Enumeration
    """
    user = 1
    slice = 2
    sliver = 3
    resources = 4
    delegation = 5


class ActionId(Enum):
    """
    Action Id Enumeration
    """
    query = 1
    status = 2
    create = 3
    redeem = 4
    delete = 5
    modify = 6
    POA = 7
    renew = 8
    demand = 9
    update = 10
    close = 11
    claim = 12
    reclaim = 13
    ticket = 14
    extend = 15
    relinquish = 16


class PdpAuth:
    """
    Responsible for Authorization against PDP
    """
    co_manage_project_leads_project = 'project-leads'
    project_lead_role = 'projectLead'

    request = 'Request'
    attribute = 'Attribute'
    attribute_id = 'AttributeId'
    category = 'Category'
    category_id = 'CategoryId'
    value = 'Value'
    email = 'email'
    subject_id_urn = 'urn:oasis:names:tc:xacml:1.0:subject:subject-id'
    category_subject_urn = 'urn:oasis:names:tc:xacml:1.0:subject-category:access-subject'
    category_resource_urn = 'urn:oasis:names:tc:xacml:3.0:attribute-category:resource'
    resource_type_urn = 'urn:fabric:xacml:attributes:resource-type'
    resource_id_urn = 'urn:oasis:names:tc:xacml:1.0:resource:resource-id'
    category_action_urn = 'urn:oasis:names:tc:xacml:3.0:attribute-category:action'
    action_id_urn = 'urn:oasis:names:tc:xacml:1.0:action:action-id'
    category_environment_urn = 'urn:oasis:names:tc:xacml:3.0:attribute-category:environment'

    missing_parameter = "Missing {}"

    subject_fabric_role_attribute_json = {
        "IncludeInResult": False,
        "AttributeId": "urn:fabric:xacml:attributes:fabric-role",
        "DataType": "http://www.w3.org/2001/XMLSchema#string",
        "Value": ["projectMember:project-X"]
    }

    resource_id_attribute_json = {
        "IncludeInResult": False,
        "AttributeId": "urn:oasis:names:tc:xacml:1.0:resource:resource-id",
        "DataType": "http://www.w3.org/2001/XMLSchema#string",
        "Value": ["some-delegation"]
    }

    def __init__(self, *, config: dict, logger=None):
        self.roles_re = 'CO:COU:(.*):members:active'
        self.project_member = "projectMember:{}"
        self.config = config
        self.logger = logger

    @staticmethod
    def _headers() -> dict:
        """
        Returns Headers for REST APIs
        """
        headers = {
            'Content-Type': "application/json"
        }
        return headers

    def get_roles(self, *, fabric_token: dict) -> List[str]:
        """
        Get Roles from a fabric token
        @param fabric_token fabric token
        @return list of the roles
        """
        ret_val = []
        roles = fabric_token.get('roles', None)
        if roles is None:
            raise PdpAuthException(self.missing_parameter.format("roles"))

        return roles

    def update_subject_category(self, *, subject: dict, token: dict) -> dict:
        """
        Update the Subject Category in PDP request
        @param subject subject
        @param token fabric token
        @return updated subject category
        """
        attributes = subject.get(PdpAuth.attribute, None)
        if attributes is None:
            raise PdpAuthException(self.missing_parameter.format("attributes"))

        roles = self.get_roles(fabric_token=token)

        if len(attributes) > 1:
            raise PdpAuthException("Should only have subject Id Attribute {}".format(subject))

        if attributes[0][PdpAuth.attribute_id] != PdpAuth.subject_id_urn:
            raise PdpAuthException("Should only have subject Id Attribute {}".format(subject))

        attributes[0][PdpAuth.value] = [token[PdpAuth.email]]

        if len(roles) < 1:
            raise PdpAuthException("No roles available in Token")

        for r in roles:
            if r != PdpAuth.co_manage_project_leads_project:
                attr = self.subject_fabric_role_attribute_json.copy()
                attr['Value'] = self.project_member.format(r)
                attributes.append(attr)
            else:
                attr = self.subject_fabric_role_attribute_json.copy()
                attr['Value'] = "projectLead"
                attributes.append(attr)

        return subject

    def update_resource_category(self, *, resource: dict, resource_type: ResourceType, resource_id: str = None) -> dict:
        """
        Update the Resource Category in PDP request
        @param resource resource
        @param resource_type resource type
        @param resource_id resource id
        @return updated Resource category
        """
        attributes = resource.get(PdpAuth.attribute, None)
        if attributes is None:
            raise PdpAuthException(self.missing_parameter.format("attributes"))

        if len(attributes) > 1:
            raise PdpAuthException("Should only have Resource Type Attribute {}".format(resource))

        if attributes[0][PdpAuth.attribute_id] != PdpAuth.resource_type_urn:
            raise PdpAuthException("Should only have Resource Type Attribute {}".format(resource))

        attributes[0][PdpAuth.value] = [resource_type.name]

        if resource_id is not None:
            attr = self.resource_id_attribute_json.copy()
            attr[PdpAuth.value] = resource_id
            attributes.append(attr)

        return resource

    def update_action_category(self, *, action: dict, action_id: ActionId) -> dict:
        """
        Update the Action Category in PDP request
        @param action action
        @param action_id action id
        @return updated Action category
        """
        attributes = action.get(PdpAuth.attribute, None)
        if attributes is None:
            raise PdpAuthException(self.missing_parameter.format("attributes"))

        if len(attributes) > 1:
            raise PdpAuthException("Should only have Action-Id Attribute {}".format(action))

        if attributes[0][PdpAuth.attribute_id] != PdpAuth.action_id_urn:
            raise PdpAuthException("Should only have Action-Id Attribute {}".format(action))

        attributes[0][PdpAuth.value] = [action_id.name]

        return action

    def build_pdp_request(self, *, fabric_token: dict, actor_type: ActorType,
                          action_id: ActionId, resource_type: ResourceType,
                          resource_id: str = None) -> dict:
        """
        Build PDP Request
        @param fabric_token fabric token
        @param actor_type action type
        @param action_id Action id
        @param resource_type resource_type
        @param resource_id resource_id
        @return PDP request
        """
        request_file = None
        if actor_type == ActorType.Orchestrator:
            request_file = os.path.dirname(__file__) + '/data/orchestrator-request.json'
        elif actor_type == ActorType.Broker:
            request_file = os.path.dirname(__file__) + '/data/broker-request.json'
        elif actor_type == ActorType.Authority:
            request_file = os.path.dirname(__file__) + '/data/am-request.json'
        else:
            raise PdpAuthException("Invalid Actor Type: {}".format(actor_type))

        request_json = None
        with open(request_file) as f:
            request_json = json.load(f)
            f.close()

        ## Subject
        categories = request_json[PdpAuth.request][PdpAuth.category]
        for c in categories:
            if c[PdpAuth.category_id] == PdpAuth.category_subject_urn:
                c = self.update_subject_category(subject=c, token=fabric_token)

            elif c[PdpAuth.category_id] == PdpAuth.category_resource_urn:
                c = self.update_resource_category(resource=c, resource_type=resource_type, resource_id=resource_id)

            elif c[PdpAuth.category_id] == PdpAuth.category_action_urn:
                c = self.update_action_category(action=c, action_id=action_id)

            elif c[PdpAuth.category_id] == PdpAuth.category_environment_urn:
                if self.logger is None:
                    print("Do nothing, ignore Environment category")
                else:
                    self.logger.debug("Do nothing, ignore Environment category")

            else:
                raise PdpAuthException("Invalid Category: {}".format(c))

        request_json[PdpAuth.request][PdpAuth.category] = categories

        return request_json

    def check_access(self, *, fabric_token: dict, actor_type: ActorType,
                     action_id: ActionId, resource_type: ResourceType,
                     resource_id: str = None) -> bool:
        """
        Check Access
        @param fabric_token fabric token
        @param actor_type actor type
        @param action_id action id
        @param resource_type resource type
        @param resource_id resource id
        @return true for success and failure otherwise
        @raises PdpAuthException in case of denied access or failure
        """

        pdp_request = self.build_pdp_request(fabric_token=fabric_token, actor_type=actor_type,
                                             action_id=action_id, resource_type=resource_type, resource_id=resource_id)

        self.logger.debug("PDP Auth Request: {}".format(pdp_request))

        response = requests.post(url=self.config['url'], headers=self._headers(), json=pdp_request)

        if response.status_code != 200:
            raise PdpAuthException('Authorization check failure: {}'.format(response))

        if response.json()["Response"][0]["Decision"] == "Permit":
            if self.logger is not None:
                self.logger.debug("PDP response: {}".format(response.json()))
            return True
        else:
            if self.logger is not None:
                self.logger.debug("PDP response: {}".format(response.json()))
            raise PdpAuthException('Authorization check failure: {}'.format(response.json()))


if __name__ == '__main__':
    token = {"email": "kthare10@email.unc.edu",
             "given_name": "Komal",
             "family_name": "Thareja", "name": "Komal Thareja",
             "iss": "https://cilogon.org", "sub": "http://cilogon.org/serverA/users/11904101",
             "aud": "cilogon:/client_id/1253defc60a323fcaa3b449326476099",
             "token_id": "https://cilogon.org/oauth2/idToken/156747336e2a3fbc1d66cc8fe1571d91/1603986888019",
             "auth_time": "1603986887", "exp": 1603990493, "iat": 1603986893,
             "roles": ["CO:members:active", "CO:COU:Jupyterhub:members:active", "CO:COU:project-leads:members:active"],
             "scope": "all", "project": "all"}

    config = {'url': 'http://localhost:8080/services/pdp'}
    pdp = PdpAuth(config=config)
    RESULT = pdp.check_access(fabric_token=token, actor_type=ActorType.Orchestrator,
                              action_id=ActionId.query, resource_type=ResourceType.resources)
    print(RESULT)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import traceback
from typing import TYPE_CHECKING, List

from fabric.actor.boot.configuration_exception import ConfigurationException
from fabric.actor.boot.inventory.pool_creator import PoolCreator
from fabric.actor.core.apis.i_authority import IAuthority
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.resource_pool_attribute_descriptor import ResourcePoolAttributeDescriptor, \
    ResourcePoolAttributeType
from fabric.actor.core.common.resource_pool_descriptor import ResourcePoolDescriptor
from fabric.actor.core.container.remote_actor_cache import RemoteActorCacheSingleton, RemoteActorCache
from fabric.actor.core.core.authority import Authority
from fabric.actor.core.core.broker import Broker
from fabric.actor.core.core.controller import Controller
from fabric.actor.core.delegation.simple_resource_ticket_factory import SimpleResourceTicketFactory
from fabric.actor.core.kernel.slice_factory import SliceFactory
from fabric.actor.core.manage.management_utils import ManagementUtils
from fabric.actor.core.manage.messages.client_mng import ClientMng
from fabric.actor.core.plugins.base_plugin import BasePlugin
from fabric.actor.core.plugins.config.config import Config
from fabric.actor.core.plugins.db.server_actor_database import ServerActorDatabase
from fabric.actor.core.plugins.substrate.authority_substrate import AuthoritySubstrate
from fabric.actor.core.plugins.substrate.substrate import Substrate
from fabric.actor.core.plugins.substrate.db.substrate_actor_database import SubstrateActorDatabase
from fabric.actor.core.policy.authority_calendar_policy import AuthorityCalendarPolicy
from fabric.actor.core.policy.broker_simpler_units_policy import BrokerSimplerUnitsPolicy
from fabric.actor.core.policy.controller_simple_policy import ControllerSimplePolicy
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.reflection_utils import ReflectionUtils
from fabric.actor.core.util.resource_type import ResourceType
from fabric.actor.core.core.actor import Actor
from fabric.actor.security.auth_token import AuthToken
from fabric.actor.core.apis.i_actor import ActorType

if TYPE_CHECKING:
    from fabric.actor.boot.configuration import Configuration, ActorConfig, PolicyConfig, Peer
    from fabric.actor.core.apis.i_actor import IActor
    from fabric.actor.core.apis.i_mgmt_actor import IMgmtActor


class ExportAdvertisement:
    """
    Encapsulates information needed for the delegations to be exported which are later claimed by broker
    """
    def __init__(self, *, exporter: IMgmtActor, client: ClientMng, delegation: str, topic: str):
        self.exporter = exporter
        self.client = client
        self.delegation = delegation
        self.client_topic = topic
        self.exported = None


class ConfigurationProcessor:
    """
    Processing the Configuration loaded and brings up the required actors, peers and various threads
    """
    def __init__(self, *, config: Configuration):
        from fabric.actor.core.container.globals import GlobalsSingleton
        self.logger = GlobalsSingleton.get().get_logger()
        self.config = config
        self.actor = None
        self.to_export = []
        self.to_advertise = []
        self.pools = {}
        self.resources = None
        self.aggregate_delegation_models = None

    def process(self):
        """
        Instantiates the configuration
        @raises ConfigurationException in case of error
        """
        try:
            self.create_actor()
            self.initialize_actor()
            self.logger.info("There are {} actors".format(Actor.actor_count))
            self.register_actor()
            self.create_default_slice()
            self.populate_inventory_neo4j()
            self.populate_inventory()
            self.recover_actor()
            self.enable_ticking()
            self.process_topology()
            self.logger.info("Processing exports with actor count {}".format(Actor.actor_count))
            self.process_advertise()
            self.logger.info("Processing exports completed")
        except Exception as e:
            self.logger.error(traceback.format_exc())
            raise ConfigurationException("Unexpected error while processing configuration {}".format(e))
        self.logger.info("Finished instantiating actors.")
        print("End process")

    def create_actor(self):
        """
        Instantiates all actors
        @raises ConfigurationException in case of error
        """
        try:
            if self.config.get_actor() is not None:
                self.logger.debug("Creating Actor: name={}".format(self.config.get_actor().get_name()))
                self.actor = self.do_common(actor_config=self.config.get_actor())
                self.do_specific(actor=self.actor, config=self.config.get_actor())
        except Exception as e:
            raise ConfigurationException("Unexpected error while creating actor {}".format(e))

    def do_common(self, *, actor_config: ActorConfig):
        """
        Perform the common operations for actor creation
        @param actor_config actor config
        @raises ConfigurationException in case of error
        """
        actor = self.make_actor_instance(actor_config=actor_config)
        actor.set_plugin(plugin=self.make_plugin_instance(actor=actor, actor_config=actor_config))
        actor.set_policy(policy=self.make_actor_policy(actor=actor, config=actor_config))
        return actor

    @staticmethod
    def make_actor_instance(*, actor_config: ActorConfig) -> IActor:
        """
        Creates Actor instance
        @param actor_config actor config
        @raises ConfigurationException in case of error
        """
        actor_type = ActorType.get_actor_type_from_string(actor_type=actor_config.get_type())
        actor = None
        if actor_type == ActorType.Orchestrator:
            actor = Controller()
        elif actor_type == ActorType.Broker:
            actor = Broker()
        elif actor_type == ActorType.Authority:
            actor = Authority()
        else:
            raise ConfigurationException("Unsupported actor type: {}".format(actor_type))

        actor_guid = ID()
        if actor_config.get_guid() is not None:
            actor_guid = ID(uid=actor_config.get_guid())
        auth_token = AuthToken(name=actor_config.get_name(), guid=actor_guid)
        actor.set_identity(token=auth_token)
        if actor_config.get_description() is not None:
            actor.set_description(description=actor_config.get_description())

        from fabric.actor.core.container.globals import GlobalsSingleton
        actor.set_actor_clock(clock=GlobalsSingleton.get().get_container().get_actor_clock())

        return actor

    def make_plugin_instance(self, *, actor: IActor, actor_config: ActorConfig):
        """
        Creates Plugin instance for the Actor
        @param actor actor
        @param actor_config actor config
        @raises ConfigurationException in case of error
        """
        plugin = None
        if actor.get_plugin() is None:
            if actor.get_type() == ActorType.Authority:
                # TODO replacement of ANT CONFIG
                plugin = AuthoritySubstrate(actor=actor, db=None, config=Config())

            elif actor.get_type() == ActorType.Orchestrator:
                # TODO replacement of ANT CONFIG
                plugin = Substrate(actor=actor, db=None, config=Config())

            elif actor.get_type() == ActorType.Broker:
                plugin = BasePlugin(actor=actor, db=None, config=Config())

        if plugin is None:
            raise ConfigurationException("Cannot instantiate plugin for actor: {}".format(
                actor_config.get_name()))

        if plugin.get_database() is None:
            db = None
            user = self.config.get_global_config().get_database()[Constants.property_conf_db_user]
            password = self.config.get_global_config().get_database()[Constants.property_conf_db_password]
            db_host = self.config.get_global_config().get_database()[Constants.property_conf_db_host]
            db_name = self.config.get_global_config().get_database()[Constants.property_conf_db_name]
            if isinstance(plugin, Substrate):
                db = SubstrateActorDatabase(user=user, password=password, database=db_name, db_host=db_host,
                                            logger=self.logger)
            else:
                db = ServerActorDatabase(user=user, password=password, database=db_name, db_host=db_host,
                                         logger=self.logger)

            plugin.set_database(db=db)

        ticket_factory = SimpleResourceTicketFactory()
        ticket_factory.set_actor(actor=actor)
        plugin.set_ticket_factory(ticket_factory=ticket_factory)
        return plugin

    def make_actor_policy(self, *, actor: IActor, config: ActorConfig):
        """
        Creates Actor Policy instance
        @param actor actor
        @param actor_config actor config
        @raises ConfigurationException in case of error
        """
        policy = None
        if config.get_policy() is None:
            if actor.get_type() == ActorType.Authority:
                policy = self.make_site_policy(config=config)
            elif actor.get_type() == ActorType.Broker:
                policy = BrokerSimplerUnitsPolicy()
            elif actor.get_type() == ActorType.Orchestrator:
                policy = ControllerSimplePolicy()
        else:
            policy = self.make_policy(policy=config.get_policy())

        if policy is None:
            raise ConfigurationException("Could not instantiate policy for actor: {}".format(config.get_name()))
        return policy

    def make_site_policy(self, *, config: ActorConfig):
        """
        Creates AM Policy instance and set up controls
        @param config actor config
        @raises ConfigurationException in case of error
        """
        # TODO KOMAL
        #if config.get_controls() is None or len(config.get_controls()) == 0:
        #    raise ConfigurationException("Missing authority policy but no control has been specified")

        policy = AuthorityCalendarPolicy()
        for c in config.get_controls():
            try:
                if c.get_module_name() is None or c.get_class_name() is None:
                    raise ConfigurationException("Missing control class name")

                control = ReflectionUtils.create_instance(module_name=c.get_module_name(),
                                                          class_name=c.get_class_name())

                if c.get_type() is None:
                    raise ConfigurationException("No type specified for control")

                control.add_type(rtype=ResourceType(resource_type=c.get_type()))
                policy.register_control(control=control)
            except Exception as e:
                traceback.print_exc()
                raise ConfigurationException("Could not create control {}".format(e))
        return policy

    @staticmethod
    def make_policy(*, policy: PolicyConfig):
        """
        Creates Policy Instance
        @param policy policy config
        @raises ConfigurationException in case of error
        """
        if policy.get_class_name() is None or policy.get_module_name() is None:
            raise ConfigurationException("Policy is missing class name")

        return ReflectionUtils.create_instance(module_name=policy.get_module_name(), class_name=policy.get_class_name())

    def do_specific(self, *, actor: IActor, config: ActorConfig):
        """
        Do Actor specify intialization
        @param actor actor
        @param config actor config
        @raises ConfigurationException in case of error
        """
        if isinstance(actor, IAuthority):
            self.pools = self.read_resource_pools(config=config)
            self.resources = self.read_resource_pools2(config=config)

    def read_resource_pools(self, *, config: ActorConfig) -> dict:
        """
        Read Resource pools and create pools
        @param config actor config
        @raises ConfigurationException in case of error
        """
        result = {}
        pools = config.get_pools()
        if pools is None or len(pools) == 0:
            return result

        for p in pools:
            descriptor = ResourcePoolDescriptor()
            descriptor.set_resource_type(rtype=ResourceType(resource_type=p.get_type()))
            descriptor.set_resource_type_label(rtype_label=p.get_label())
            descriptor.set_units(units=p.get_units())
            descriptor.set_start(start=p.get_start())
            descriptor.set_end(end=p.get_end())
            descriptor.set_pool_factory_class(factory_class=p.get_factory_class())
            descriptor.set_pool_factory_module(factory_module=p.get_factory_module())
            handler = p.get_handler()
            if handler is not None:
                descriptor.set_handler_class(handler_class=handler.get_class_name())
                descriptor.set_handler_module(module=handler.get_module_name())
                descriptor.set_handler_properties(properties=handler.get_properties())

            descriptor.pool_properties = p.get_properties()

            for attr in p.get_attributes():
                attribute = ResourcePoolAttributeDescriptor()
                attribute.set_key(value=attr.get_key())
                attribute.set_value(value=attr.get_value())
                if attr.get_type().lower() == "integer":
                    attribute.set_type(rtype=ResourcePoolAttributeType.INTEGER)
                elif attr.get_type().lower() == "string":
                    attribute.set_type(rtype=ResourcePoolAttributeType.STRING)
                elif attr.get_type().lower() == "neo4j":
                    attribute.set_type(rtype=ResourcePoolAttributeType.NEO4J)
                elif attr.get_type().lower() == "class":
                    attribute.set_type(rtype=ResourcePoolAttributeType.CLASS)
                else:
                    raise ConfigurationException("Unsupported attribute type: {}".format(attr.get_type()))
                descriptor.add_attribute(attribute=attribute)

            result[descriptor.get_resource_type()] = descriptor
        return result

    def read_resource_pools2(self, *, config: ActorConfig) -> dict:
        """
        Read resource pool config and create pools
        @param config actor config
        @raises ConfigurationException in case of error
        """
        result = {}
        resources = config.get_resources()
        if resources is None or len(resources) == 0:
            return result

        for r in resources:
            descriptor = ResourcePoolDescriptor()
            descriptor.set_resource_type(rtype=ResourceType(resource_type=r.get_type()))
            descriptor.set_resource_type_label(rtype_label=r.get_label())

            # TODO for now it holds the control name
            descriptor.set_pool_factory_class(factory_class=r.get_control().get_class_name())
            descriptor.set_pool_factory_module(factory_module=r.get_control().get_module_name())

            handler = r.get_handler()
            if handler is not None:
                descriptor.set_handler_class(handler_class=handler.get_class_name())
                descriptor.set_handler_module(module=handler.get_module_name())
                descriptor.set_handler_properties(properties=handler.get_properties())

            descriptor.pool_properties = r.get_properties()

            for attr in r.get_attributes():
                attribute = ResourcePoolAttributeDescriptor()
                attribute.set_key(value=attr.get_key())
                attribute.set_value(value=attr.get_value())
                if attr.get_type().lower() == "integer":
                    attribute.set_type(rtype=ResourcePoolAttributeType.INTEGER)
                elif attr.get_type().lower() == "string":
                    attribute.set_type(rtype=ResourcePoolAttributeType.STRING)
                elif attr.get_type().lower() == "neo4j":
                    attribute.set_type(rtype=ResourcePoolAttributeType.NEO4J)
                elif attr.get_type().lower() == "class":
                    attribute.set_type(rtype=ResourcePoolAttributeType.CLASS)
                else:
                    raise ConfigurationException("Unsupported attribute type: {}".format(attr.get_type()))
                descriptor.add_attribute(attribute=attribute)

            result[descriptor.get_resource_type()] = descriptor
        return result

    def initialize_actor(self):
        """
        Initialize actor
        @raises ConfigurationException in case of error
        """
        try:
            Actor.actor_count += 1
            self.actor.initialize()
        except Exception as e:
            raise ConfigurationException("Actor failed to initialize: {} {}".format(self.actor.get_name(), e))

    def register_actor(self):
        """
        Register actor
        @raises ConfigurationException in case of error
        """
        try:
            from fabric.actor.core.container.globals import GlobalsSingleton
            GlobalsSingleton.get().get_container().register_actor(actor=self.actor)
        except Exception as e:
            raise ConfigurationException("Could not register actor: {} {}".format(self.actor.get_name(), e))

    def create_default_slice(self):
        """
        Create default slice
        @raises ConfigurationException in case of error
        """
        if self.actor.get_type() != ActorType.Authority:
            slice_obj = SliceFactory.create(slice_id=ID(), name=self.actor.get_name())
            slice_obj.set_inventory(value=True)
            try:
                self.actor.register_slice(slice_object=slice_obj)
            except Exception as e:
                traceback.print_exc()
                raise ConfigurationException("Could not create default slice for actor: {} {}".format(
                    self.actor.get_name(), e))

    def populate_inventory(self):
        """
        Populate inventory
        @raises ConfigurationException in case of error
        """
        if isinstance(self.actor, IAuthority) and isinstance(self.actor.get_plugin(), AuthoritySubstrate):
            creator = PoolCreator(substrate=self.actor.get_plugin(), pools=self.pools,
                                  neo4j_config=self.config.get_global_config().get_neo4j_config())
            creator.process()

    def populate_inventory_neo4j(self):
        """
        Load Aggregate Resource model
        @raises ConfigurationException in case of error
        """
        if isinstance(self.actor, IAuthority) and isinstance(self.actor.get_plugin(), AuthoritySubstrate):
            creator = PoolCreator(substrate=self.actor.get_plugin(), pools=self.resources,
                                  neo4j_config=self.config.get_global_config().get_neo4j_config())
            self.aggregate_delegation_models = creator.process_neo4j(actor_name=self.actor.get_name(),
                                                                     substrate_file=self.config.get_actor().
                                                                     get_substrate_file())

    def recover_actor(self):
        """
        Recover an actor on stateful restart
        @raises ConfigurationException in case of error
        """
        try:
            self.actor.recover()
        except Exception as e:
            raise ConfigurationException("Recovery failed for actor: {} e: {}".format(self.actor.get_name(), e))

    def enable_ticking(self):
        """
        Enable the Actor tick
        """
        from fabric.actor.core.container.globals import GlobalsSingleton
        GlobalsSingleton.get().get_container().register(tickable=self.actor)

    def process_topology(self):
        """
        Set up Peer registry with identity and kafka topic information
        @raises ConfigurationException in case of error
        """
        if self.config.get_peers() is None or len(self.config.get_peers()) == 0:
            self.logger.debug("No peers specified")
        else:
            self.create_proxies(peers=self.config.get_peers())

    def process_advertise(self):
        """
        Set up Aggregate Delegation models for the peers
        @raises ConfigurationException in case of error
        """
        if self.aggregate_delegation_models is None and len(self.to_advertise) > 0:
            raise ConfigurationException("No delegations found in Aggregate Resource Model")

        for ei in self.to_advertise:
            self.advertise(info=ei)

    def create_proxies(self, *, peers: List[Peer]):
        """
        Set up proxies
        @param peers List of the Peer config
        @raises ConfigurationException in case of error
        """
        for e in peers:
            self.process_peer(peer=e)

    def vertex_to_registry_cache(self, *, peer: Peer):
        """
        Add peer to Registry Cache
        @param peer peer config
        @raises ConfigurationException in case of error
        """
        self.logger.debug("Adding vertex for {}".format(peer.get_name()))

        if peer.get_name() is None:
            raise ConfigurationException("Actor must specify a name")

        if peer.get_guid() is None:
            raise ConfigurationException("Actor must specify a guid")

        protocol = Constants.protocol_local
        kafka_topic = None
        if peer.get_kafka_topic() is not None:
            protocol = Constants.protocol_kafka
            kafka_topic = peer.get_kafka_topic()

        actor_type = peer.get_type().lower()

        entry = {
            RemoteActorCache.actor_name: peer.get_name(),
            RemoteActorCache.actor_guid: ID(uid=peer.get_guid()),
            RemoteActorCache.actor_type: actor_type,
            RemoteActorCache.actor_protocol: protocol
        }
        if kafka_topic is not None:
            entry[RemoteActorCache.actor_location] = kafka_topic

        RemoteActorCacheSingleton.get().add_partial_cache_entry(guid=ID(uid=peer.get_guid()), entry=entry)

    def process_peer(self, *, peer: Peer):
        """
        Process a peer
        @param peer peer
        @raises ConfigurationException in case of error
        """
        from_guid = ID(uid=peer.get_guid())
        from_type = ActorType.get_actor_type_from_string(actor_type=peer.get_type())
        to_guid = self.actor.get_guid()
        to_type = self.actor.get_type()

        # We only like peers broker->site and service->broker
        # Reverse the peer if it connects site->broker or broker->service

        if from_type == ActorType.Authority and to_type == ActorType.Broker:
            from_guid, to_guid = to_guid, from_guid
            from_type, to_type = to_type, from_type

        if from_type == ActorType.Broker and to_type == ActorType.Orchestrator:
            from_guid, to_guid = to_guid, from_guid
            from_type, to_type = to_type, from_type

        # peers between actors of same type aren't allowed unless the actors are both brokers
        if from_type == to_type and from_type != ActorType.Broker:
            raise ConfigurationException(
                "Invalid peer type: broker can only talk to broker, orchestrator or site authority")

        container = ManagementUtils.connect(caller=self.actor.get_identity())
        to_mgmt_actor = container.get_actor(guid=to_guid)
        if to_mgmt_actor is None:
            self.logger.debug("to_mgmt_actor={} to_guid={}".format(type(to_mgmt_actor), to_guid))
            if container.get_last_error() is not None:
                self.logger.error(container.get_last_error())
        from_mgmt_actor = container.get_actor(guid=from_guid)
        if from_mgmt_actor is None:
            self.logger.debug("from_mgmt_actor={} from_guid={}".format(type(from_mgmt_actor), from_guid))
            if container.get_last_error() is not None:
                self.logger.error(container.get_last_error())

        self.vertex_to_registry_cache(peer=peer)

        try:
            client = RemoteActorCacheSingleton.get().establish_peer(from_guid=from_guid,
                                                                    from_mgmt_actor=from_mgmt_actor,
                                                                    to_guid=to_guid, to_mgmt_actor=to_mgmt_actor)
            self.logger.debug("Client returned {}".format(client))
            if client is not None:
                self.parse_exports(peer=peer, client=client, mgmt_actor=to_mgmt_actor)
        except Exception as e:
            raise ConfigurationException("Could not process exports from: {} to {}. e= {}" .format(
                peer.get_guid(), self.actor.get_guid(), e))

    def parse_exports(self, *, peer: Peer, client: ClientMng, mgmt_actor: IMgmtActor):
        """
        Identify all the delegations to be advertised
        @param peer peer config
        @param client client
        @param mgmt_actor management actor
        """
        if peer.get_delegation() is not None:
            info = ExportAdvertisement(exporter=mgmt_actor, client=client,
                                       delegation=peer.get_delegation(),
                                       topic=peer.get_kafka_topic())
            self.to_advertise.append(info)

    def advertise(self, *, info: ExportAdvertisement):
        """
        Advertise a delegation
        @param info advertisement information
        @raises ConfigurationException in case of error
        """
        self.logger.debug("Using Server Actor {} to export resources".format(info.exporter.__class__.__name__))

        delegation = self.aggregate_delegation_models.get(info.delegation, None)

        if delegation is None:
            raise ConfigurationException("Aggregate Delegation Model not found for delegation: {}".format(
                info.delegation))

        info.exported = info.exporter.advertise_resources(delegation=delegation,
                                                          client=AuthToken(name=info.client.get_name(),
                                                                           guid=ID(uid=info.client.get_guid())))

        if info.exported is None:
            raise ConfigurationException("Could not export resources from actor: {} to actor: {} Error = {}".format(
                info.exporter.get_name(), info.client.get_name(), info.exporter.get_last_error()))



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.apis.i_state import IState
from fabric.actor.core.kernel.reservation_states import JoinState, ReservationPendingStates, ReservationStates


class ReservationState(IState):
    """
    Represents the state of a reservation.
    """
    def __init__(self, *, state: ReservationStates, pending: ReservationPendingStates, joining: JoinState = -1):
        self.state = state
        self.pending = pending
        self.joining = joining

    def __eq__(self, other):
        if other is None:
            return False

        if not isinstance(other, ReservationState):
            return NotImplemented

        return self.state == other.state and self.pending == other.pending and self.joining == other.joining

    def get_joining(self) -> JoinState:
        """
        Returns the joining sub-state.
      
        @return joining sub-state
        """
        return self.joining

    def get_joining_name(self) -> str:
        """
        Returns the joining sub-state name.

        @return joining sub-state name
        """
        return self.joining.name

    def get_pending(self) -> ReservationPendingStates:
        """
        Returns the pending sub-state.

        @return pending sub-state
        """
        return self.pending

    def get_pending_name(self) -> str:
        """
        Returns the pending sub-state name.

        @return pending sub-state name
        """
        return self.pending.name

    def get_state(self) -> ReservationStates:
        """
        Returns the state.

        @return state
        """
        return self.state

    def get_state_name(self) -> str:
        """
        Returns the state name

        @return state name
        """
        return self.state.name

    def __str__(self):
        if self.joining == -1:
            return "[{}, {}]".format(self.get_state_name(), self.get_pending_name())
        else:
            return "[{}, {}, {}]".format(self.get_state_name(), self.get_pending_name(), self.get_joining_name())



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)

from fabric.actor.core.apis.i_actor import IActor
from fabric.actor.core.common.exceptions import ResourcesException
from fabric.actor.core.common.resource_vector import ResourceVector
from fabric.actor.core.apis.i_resource_ticket_factory import IResourceTicketFactory
from fabric.actor.core.delegation.resource_bin import ResourceBin
from fabric.actor.core.delegation.resource_delegation import ResourceDelegation
from fabric.actor.core.delegation.resource_ticket import ResourceTicket
from fabric.actor.core.time.actor_clock import ActorClock
from fabric.actor.core.time.term import Term
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.resource_type import ResourceType


class SimpleResourceTicketFactory(IResourceTicketFactory):
    """
    Factory class to create Resource Tickets
    """
    def __init__(self):
        self.actor = None
        self.initialized = False

    def initialize(self):
        """
        Initialize
        """
        if not self.initialized:
            if self.actor is None:
                raise ResourcesException("Factory does not have an actor")
            self.initialized = True

    def ensure_initialized(self):
        """
        Ensure if the factory is initialized
        """

        if not self.initialized:
            raise ResourcesException("ticket factory has not been initialized")

    def get_issuer_id(self):
        """
        Get Actor Guid
        """
        return self.actor.get_identity().get_guid()

    def make_delegation(self, *, units: int = None, vector: ResourceVector = None, term: Term = None,
                        rtype: ResourceType = None, sources: list = None, bins: list = None,
                        properties: dict = None, holder: ID = None) -> ResourceDelegation:
        self.ensure_initialized()
        if (sources is None and bins is not None) or (sources is not None and bins is None):
            raise ResourcesException("sources and bins must both be null or non-null")

        issuer = self.get_issuer_id()

        return ResourceDelegation(units=units, vector=vector, term=term, rtype=rtype, sources=sources, bins=bins,
                                  properties=properties, issuer=issuer, holder=holder)

    def make_ticket(self, *, delegation: ResourceDelegation = None, source: ResourceTicket = None) -> ResourceTicket:
        self.ensure_initialized()
        return ResourceTicket(factory=self, delegation=delegation, source=source)

    def set_actor(self, *, actor: IActor):
        self.actor = actor

    def get_actor(self) -> IActor:
        return self.actor

    def clone(self, *, original: ResourceTicket) -> ResourceTicket:
        self.ensure_initialized()
        ticket_json = self.to_json(ticket=original)
        result = self.from_json(incoming=ticket_json)
        return result

    def to_json(self, *, ticket: ResourceTicket) -> dict:
        outgoing_ticket = {}
        delegation_list = []
        for delegation in ticket.delegations:
            d = {'guid': delegation.get_guid()}
            t = {'start_time': ActorClock.to_milliseconds(when=delegation.get_term().get_start_time()),
                 'end_time': ActorClock.to_milliseconds(when=delegation.get_term().get_end_time()),
                 'new_start_time': ActorClock.to_milliseconds(when=delegation.get_term().get_new_start_time())}
            d['term'] = t
            d['units'] = delegation.get_units()
            d['type'] = delegation.get_resource_type()
            d['issuer'] = delegation.get_issuer()
            if delegation.get_holder():
                d['holder'] = delegation.get_holder()
            if delegation.sources is not None:
                source_list = []
                for s in delegation.sources:
                    val = {'id': s}
                    source_list.append(val)
                d['sources'] = source_list
            if delegation.bins is not None:
                bin_list = []
                for b in delegation.bins:
                    b_bin = {'guid': b.get_guid(), 'physical_units': b.get_physical_units()}

                    if b.get_parent_guid() is not None:
                        b['parent_guid'] = b.get_parent_guid()

                    term = {'start_time': ActorClock.to_milliseconds(when=b.get_term().get_start_time()),
                            'end_time': ActorClock.to_milliseconds(when=b.get_term().get_end_time()),
                            'new_start_time': ActorClock.to_milliseconds(when=b.get_term().get_new_start_time())}

                    b_bin['term'] = term
                    bin_list.append(b_bin)
                d['bins'] = bin_list

            delegation_list.append(d)
        outgoing_ticket['delegations'] = delegation_list
        return {'ticket': outgoing_ticket}

    def from_json(self, *, incoming: dict) -> ResourceTicket:
        ticket = None

        incoming_ticket = incoming.get('ticket', None)
        if incoming_ticket is not None:
            delegations = incoming_ticket.get('delegations', None)
            if delegations is not None:
                for d in delegations:
                    dd = ResourceDelegation()
                    dd.guid = d.get('guid', None)
                    term_dict = d.get('term', None)
                    if term_dict is not None:
                        start_time = term_dict.get('start_time', None)
                        end_time = term_dict.get('end_time', None)
                        new_start_time = term_dict.get('new_start_time', None)
                        term = Term(start=ActorClock.from_milliseconds(milli_seconds=start_time),
                                    end=ActorClock.from_milliseconds(milli_seconds=end_time),
                                    new_start=ActorClock.from_milliseconds(milli_seconds=new_start_time))
                        dd.term = term
                    dd.units = d.get('units', 0)
                    dd.type = d.get('type', None)
                    dd.issuer = d.get('issuer', None)
                    dd.holder = d.get('holder', None)
                    sources = d.get('sources', None)
                    if sources is not None:
                        dd.sources = []
                        for s in sources:
                            if s.get('id', None) is not None:
                                dd.sources.append(s['id'])
                    bins = d.get('bins', None)
                    if bins is not None:
                        dd.bins = []
                        for b in bins:
                            bb = ResourceBin()
                            bb.guid = b.get('guid', None)
                            bb.physical_units = b.get('physical_units', 0)
                            bb.parent_guid = b.get('parent_guid', None)
                            term_dict = b.get('term', None)
                            if term_dict is not None:
                                start_time = term_dict.get('start_time', None)
                                end_time = term_dict.get('end_time', None)
                                new_start_time = term_dict.get('new_start_time', None)
                                term = Term(start=ActorClock.from_milliseconds(milli_seconds=start_time),
                                            end=ActorClock.from_milliseconds(milli_seconds=end_time),
                                            new_start=ActorClock.from_milliseconds(milli_seconds=new_start_time))
                                bb.term = term
                            dd.bins.append(bb)
                    if ticket is None:
                        ticket = ResourceTicket(factory=self, delegation=dd, source=None)
                    else:
                        ticket.delegations.append(dd)

        return ticket



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING

from fabric.actor.core.apis.i_authority_reservation_factory import IAuthorityReservationFactory
from fabric.actor.core.util.id import ID

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_slice import ISlice
    from fabric.actor.core.kernel.resource_set import ResourceSet
    from fabric.actor.core.time.term import Term


class AuthorityReservationFactory(IAuthorityReservationFactory):
    """
    Factory class for creating authority reservations
    """
    @staticmethod
    def create(*, resources: ResourceSet, term: Term, slice_obj: ISlice, rid: ID):
        """
        Create authority reservation
        @param resources resources
        @param term term
        @param slice_obj slice object
        @param rid reservation id
        @return Authority Reservation
        """
        from fabric.actor.core.kernel.authority_reservation import AuthorityReservation
        res = AuthorityReservation(rid=rid, resources=resources, term=term, slice_object=slice_obj)
        return res



import json
import traceback

import jwt
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization


class TokenException(Exception):
    """
    Token exception
    """


class FabricToken:
    """
    Represents the Fabric Token issues by Credential Manager
    """
    def __init__(self, *, token_public_key: str, token: str, logger):
        if token_public_key is None or token is None:
            raise TokenException('Either token_public_key: {} or token: {} is None'.format(token_public_key, token))

        self.token_public_key = token_public_key
        self.logger = logger
        self.encoded_token = token
        self.decoded_token = None

    def get_encoded_token(self) -> str:
        """
        Get Encoded token string
        @return encoded token
        """
        return self.encoded_token

    def get_decoded_token(self) -> dict:
        """
        Get Decoded token
        @return Decoded token
        """
        return self.decoded_token

    def validate(self) -> dict:
        """
        Validate the token
        @raise Exception in case of error
        """
        try:
            with open(self.token_public_key) as f:
                pem_data = f.read()
                f.close()
                key = serialization.load_pem_public_key(data=pem_data.encode("utf-8"),
                                                        backend=default_backend())

            options = {'verify_aud': False}
            verify = True
            self.decoded_token = jwt.decode(self.encoded_token, key=key, algorithms='RS256', options=options,
                                            verify=verify)
            self.logger.debug(json.dumps(self.decoded_token))

            return self.decoded_token
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error("Exception occurred while validating the token e: {}".format(e))
            raise e



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
import ipaddress

from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import FrameworkException


class IPv4Set:
    entry_separator = ","
    subnet_mark = "/"
    range_mark = "-"

    def __init__(self, *, ip_list: str = None):
        self.free_set = set()
        self.allocated = set()

        if ip_list:
            self.add(ip_list=ip_list)

    def add(self, *, ip_list: str):
        for token in ip_list.split(self.entry_separator):
            if token.find(self.subnet_mark) > -1:
                self.process_subnet(token=token)
            elif token.find(self.range_mark) > -1:
                self.process_range(token=token)
            else:
                self.process_single(token=token)

    def process_subnet(self, *, token: str):
        tokens = token.split(self.subnet_mark)
        if len(tokens) != 2:
            raise FrameworkException("Invalid subnet: {}".find(token))

        base = self.to_ip4(ip=tokens[0])

        size = int(tokens[1])

        if size < 16 or size > 32:
            raise FrameworkException("Invalid subnet size: {}".format(size))

        network = ipaddress.IPv4Network(token, False)
        all_ips = list(network.hosts())
        last = all_ips[-1]
        start = base
        end = int(last)

        while start <= end:
            self.free_set.add(start)
            start += 1

    def process_range(self, *, token: str):
        tokens = token.split(self.range_mark)
        if len(tokens) != 2:
            raise FrameworkException("Invalid range: {}".find(token))

        start = str(self.pad_if_needed(token=tokens[0]))
        start_ip = self.to_ip4(ip=start)
        end = tokens[1]

        if end.find(".") == -1:
            end = start[0:start.rfind(".")] + "." + end

        end_ip = self.to_ip4(ip=end)

        size = end_ip - start_ip + 1

        if size < 0 | size > 65536:
            raise FrameworkException("Range must be positive and less than 65536")

        for i in range(size):
            self.free_set.add(start_ip + i)

    def pad_if_needed(self, *, token: str):
        index = 0
        count = 0
        index = token.find(".")

        while index != -1:
            count += 1
            index = token.find(".", index+1)

        for i in range(3-count):
            token = token + ".0"

        return token

    def process_single(self, *, token: str):
        self.free_set.add(self.to_ip4(ip=token))

    def allocate(self):
        item = self.free_set.pop()
        self.allocated.add(item)
        ret_val = self.int_to_ip4(ip=item)
        return ret_val

    def free(self, *, ip: str):
        val = self.to_ip4(ip=ip)
        self.allocated.remove(val)
        self.free_set.add(val)

    def reserve(self, *, ip: str):
        val = self.to_ip4(ip=ip)
        self.free_set.remove(val)
        self.allocated.add(val)

    def get_free_count(self) -> int:
        return len(self.free_set)

    def is_free(self, *, ip: int) -> bool:
        if ip in self.free_set:
            return True
        return False

    def is_allocated(self, *, ip: int) -> bool:
        if ip in self.allocated:
            return True
        return False

    def to_ip4(self, *, ip: str):
        tokens = ip.split(".")
        if len(tokens) != 4:
            raise FrameworkException(Constants.invalid_ip.format(ip))

        result = int(ipaddress.ip_address(ip))

        return result

    def int_to_ip4(self, *, ip: int):
        result = ipaddress.ip_address(ip).__str__()

        return result

    def __str__(self):
        return "IP Free:{}\nIP Allocated:{}".format(self.free_set, self.allocated)






#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.common.exceptions import ResourcesException


class ResourceData:
    """
    A ResourceData contains several collections of properties describing resources.
    Some of these collections are passed between actors during calls.
    """
    def __init__(self):
        self.local_properties = {}
        self.request_properties = {}
        self.resource_properties = {}
        self.configuration_properties = {}

    def __getstate__(self):
        return self.__dict__.copy()

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __str__(self):
        return "local: {}, request: {}, resource: {}, config: {}".format(self.local_properties, self.request_properties,
                                                                         self.resource_properties,
                                                                         self.configuration_properties)

    def clone(self):
        """
        Clone an object
        """
        obj = ResourceData()
        obj.local_properties = self.local_properties.copy()
        obj.request_properties = self.request_properties.copy()
        obj.resource_properties = self.resource_properties.copy()
        obj.configuration_properties = self.configuration_properties.copy()
        return obj

    def get_local_properties(self) -> dict:
        """
        Get local properties
        @return local properties
        """
        return self.local_properties

    def get_request_properties(self) -> dict:
        """
        Get Request properties
        @return request properties
        """
        return self.request_properties

    def get_resource_properties(self) -> dict:
        """
        Get resource properties
        @return resource properties
        """
        return self.resource_properties

    def get_configuration_properties(self) -> dict:
        """
        Get config properties
        @return config properties
        """
        return self.configuration_properties

    def merge(self, *, other):
        """
        Merge with other instance of ResourceData
        @param other other
        """
        if other is None or not isinstance(other, ResourceData):
            raise ResourcesException("Invalid object type")

        self.local_properties = self.merge_properties(from_props=other.local_properties,
                                                      to_props=self.local_properties)

        self.request_properties = self.merge_properties(from_props=other.request_properties,
                                                        to_props=self.request_properties)

        self.resource_properties = self.merge_properties(from_props=other.resource_properties,
                                                         to_props=self.resource_properties)

        self.configuration_properties = self.merge_properties(from_props=other.configuration_properties,
                                                              to_props=self.configuration_properties)

    @staticmethod
    def merge_properties(*, from_props: dict, to_props: dict) -> dict:
        """
        Merges both properties lists. Elements in from overwrite elements in two.

        @params from_props : from list
        @params to_props : to list
        """
        if from_props is None:
            return to_props

        if to_props is None:
            return from_props

        if from_props == to_props:
            return from_props

        result = {**from_props, **to_props}
        return result


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import time
import unittest
from typing import TYPE_CHECKING

from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.resource_pool_attribute_descriptor import ResourcePoolAttributeDescriptor, \
    ResourcePoolAttributeType
from fabric.actor.core.common.resource_pool_descriptor import ResourcePoolDescriptor
from fabric.actor.core.core.ticket import Ticket
from fabric.actor.core.core.unit import Unit
from fabric.actor.core.core.unit_set import UnitSet
from fabric.actor.core.kernel.client_reservation_factory import ClientReservationFactory
from fabric.actor.core.kernel.resource_set import ResourceSet
from fabric.actor.core.policy.resource_control import ResourceControl
from fabric.actor.core.policy.vm_control import VMControl
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.resource_data import ResourceData
from fabric.actor.test.core.policy.authority_calendar_policy_test import AuthorityCalendarPolicyTest

if TYPE_CHECKING:
    from fabric.actor.core.policy.authority_calendar_policy import AuthorityCalendarPolicy
    from fabric.actor.core.apis.i_actor import IActor
    from fabric.actor.core.apis.i_slice import ISlice
    from fabric.actor.core.time.term import Term
    from fabric.actor.core.util.resource_type import ResourceType
    from fabric.actor.core.apis.i_authority import IAuthority
    from fabric.actor.core.util.update_data import UpdateData
    from fabric.actor.core.apis.i_authority_reservation import IAuthorityReservation
    from fabric.actor.core.apis.i_reservation import IReservation


class VMControlTest(AuthorityCalendarPolicyTest, unittest.TestCase):
    VMMemory = 300
    VmmCapacity = 3

    from fabric.actor.core.container.globals import Globals
    Globals.config_file = Constants.test_vm_am_configuration_file

    from fabric.actor.core.container.globals import GlobalsSingleton
    GlobalsSingleton.get().start(force_fresh=True)
    while not GlobalsSingleton.get().start_completed:
        time.sleep(0.0001)

    def get_control(self, policy: AuthorityCalendarPolicy = None):
        if policy:
            return policy.controls_by_guid.values().__iter__().__next__()
        else:
            control = VMControl()
            properties = {ResourceControl.PropertyControlResourceTypes: str(self.Type)}
            control.configure(properties=properties)
            return control

    def get_pool_descriptor(self) -> ResourcePoolDescriptor:
        rd = ResourcePoolDescriptor()
        rd.set_resource_type(rtype=self.Type)
        rd.set_resource_type_label(rtype_label=self.Label)
        ad = ResourcePoolAttributeDescriptor()
        ad.set_key(value=Constants.resource_memory)
        ad.set_label(label="Memory")
        ad.set_unit(unit="MB")
        ad.set_type(rtype=ResourcePoolAttributeType.INTEGER)
        ad.set_value(value=str(self.VMMemory))
        rd.add_attribute(attribute=ad)
        return rd

    def get_source(self, units: int, rtype: ResourceType, term: Term, actor: IActor, slice_obj: ISlice):
        resources = ResourceSet(units=units, rtype=rtype, rdata=ResourceData())
        properties = resources.get_resource_properties()
        rd = self.get_pool_descriptor()
        resources.set_resource_properties(p=rd.save(properties=properties, prefix=None))

        local = resources.get_local_properties()
        local[VMControl.PropertyIPSubnet] = "255.255.255.0"
        local[VMControl.PropertyIPList] = "192.168.1.2/24"
        local[VMControl.PropertyCapacity] = self.VmmCapacity

        delegation = actor.get_plugin().get_ticket_factory().make_delegation(units=units, term=term, rtype=rtype,
                                                                             vector=None)
        ticket = actor.get_plugin().get_ticket_factory().make_ticket(delegation=delegation, source=None)
        cs = Ticket(ticket=ticket, plugin=actor.get_plugin())
        resources.set_resources(cset=cs)
        source = ClientReservationFactory.create(rid=ID(), resources=resources, term=term, slice_object=slice_obj)
        return source

    def get_donate_set(self, authority: IAuthority) -> ResourceSet:
        rset = ResourceSet(units=self.DonateUnits, rtype=self.Type, rdata=ResourceData())
        properties = rset.get_resource_properties()
        rd = self.get_pool_descriptor()
        properties = rd.save(properties=properties, prefix=None)
        rset.set_resource_properties(p=properties)
        local = rset.get_local_properties()
        local[VMControl.PropertyIPSubnet] = "255.255.255.0"
        local[VMControl.PropertyIPList] = "192.168.1.2/24"
        local[VMControl.PropertyCapacity] = self.VmmCapacity
        uset = UnitSet(plugin=authority.get_plugin())

        for i in range(self.DonateUnits):
            host = Unit(id=ID())
            host.set_property(name=Constants.unit_host_name, value="Host" + str(i))
            host.set_property(name=Constants.unit_management_ip, value="192.168.123." + str(i + 1))
            host.set_property(name=Constants.unit_control, value="192.168.123." + str(i + 1))
            uset.add_unit(u=host)

        rset.set_resources(cset=uset)

        return rset

    def check_before_donate_set(self, authority: IAuthority):
        super().check_before_donate_set(authority)
        policy = authority.get_policy()
        control = self.get_control(policy=policy)
        self.assertIsNotNone(control.inventory)
        self.assertEqual(0, len(control.inventory))

    def check_after_donate_set(self, authority: IAuthority, rset: ResourceSet):
        policy = authority.get_policy()
        control = self.get_control(policy)

        self.assertEqual(1, len(control.inventory))
        pool = control.inventory.values().__iter__().__next__()

        self.assertEqual(self.DonateUnits, pool.get_vmms_count())
        self.assertEqual(self.VmmCapacity, pool.get_capacity())
        self.assertEqual(self.Type, pool.get_type())

        self.assertIsNotNone(pool.vmms)

        for vmm in pool.get_vmm_set():
            self.assertIsNotNone(vmm.get_host())
            self.assertEqual(self.VmmCapacity, vmm.get_available())
            self.assertEqual(self.VmmCapacity, vmm.get_capacity())
            self.assertEqual(0, vmm.get_hosted_count())

    my_unit = None

    def check_incoming_lease(self, authority: IAuthority, request: IAuthorityReservation, incoming: IReservation,
                             udd: UpdateData):
        super().check_incoming_lease(authority, request, incoming, udd)
        policy = authority.get_policy()
        control = self.get_control(policy)
        rset = incoming.get_resources()
        uset = rset.get_resources()
        self.assertIsNotNone(uset)
        # TODO KOMAL
        if uset.get_units() != 0:
            print("Checking units")
            self.assertEqual(self.TicketUnits, uset.get_units())
            u = uset.get_set().values().__iter__().__next__()
            self.assertEqual(self.VMMemory, int(u.get_property(name=Constants.unit_memory)))
            self.assertIsNotNone(u.get_property(name=Constants.unit_management_ip))
            self.assertIsNotNone(u.get_property(name=Constants.unit_manage_subnet))

            vmm_host_id = u.get_parent_id()
            self.assertIsNotNone(vmm_host_id)

            pool = control.inventory.get(self.Type)
            self.assertIsNotNone(pool)

            vmm = pool.get_vmm(uid=vmm_host_id)
            self.assertIsNotNone(vmm)

            self.assertEqual(self.TicketUnits, vmm.get_hosted_count())
            self.assertEqual(self.VmmCapacity - self.TicketUnits, vmm.get_available())
            self.assertTrue(u in vmm.get_hosted_vms())
            print("Unit= {}".format(u))
            self.my_unit = u
        else:
            print("It should never enter here")
            assert (1 != 0)

    def check_incoming_close_lease(self, authority: IAuthority, request: IAuthorityReservation, incoming: IReservation,
                                   udd: UpdateData):
        super().check_incoming_close_lease(authority, request, incoming, udd)
        policy = authority.get_policy()
        control = self.get_control(policy)

        self.assertIsNotNone(self.my_unit)

        rset = incoming.get_resources()
        uset = rset.get_resources()
        self.assertIsNotNone(uset)
        self.assertEqual(0, uset.get_units())

        vmm_host_id = self.my_unit.get_parent_id()
        self.assertIsNotNone(vmm_host_id)

        pool = control.inventory.get(self.Type)
        self.assertIsNotNone(pool)

        vmm = pool.get_vmm(uid=vmm_host_id)
        self.assertIsNotNone(vmm)

        self.assertEqual(0, vmm.get_hosted_count())
        self.assertEqual(self.VmmCapacity, vmm.get_available())
        self.assertFalse(self.my_unit in vmm.get_hosted_vms())

    def check_incoming_extend_lease(self, authority: IAuthority, request: IAuthorityReservation, incoming: IReservation,
                                    update_data: UpdateData):
        super().check_incoming_extend_lease(authority, request, incoming, update_data)
        policy = authority.get_policy()
        control = self.get_control(policy)

        rset = incoming.get_resources()
        uset = rset.get_resources()
        self.assertIsNotNone(uset)
        self.assertEqual(self.TicketUnits, uset.get_units())

        u = uset.get_set().values().__iter__().__next__()
        self.assertEqual(self.VMMemory, int(u.get_property(name=Constants.unit_memory)))
        self.assertIsNotNone(u.get_property(name=Constants.unit_management_ip))
        self.assertIsNotNone(u.get_property(name=Constants.unit_manage_subnet))

        vmm_host_id = self.my_unit.get_parent_id()
        self.assertIsNotNone(vmm_host_id)

        pool = control.inventory.get(rtype=self.Type)
        self.assertIsNotNone(pool)

        vmm = pool.get_vmm(uid=vmm_host_id)
        self.assertIsNotNone(vmm)

        self.assertEqual(self.TicketUnits, vmm.get_hosted_count())
        self.assertEqual(self.VmmCapacity - self.TicketUnits, vmm.get_available())
        self.assertTrue(u in vmm.get_hosted_vms())
        print("Extend Unit= {}".format(u))
        self.assertEqual(self.my_unit, u)


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_slice import ISlice
    from fabric.actor.core.util.id import ID


class ISliceOperations:
    """
    ISliceOperations defines a common set of management operations for slices.
    This interface is implemented by each  actor.
    """
    @abstractmethod
    def get_client_slices(self) -> List[ISlice]:
        """
        Returns all client slices registered with the actor.

        Returns:
            an array of client slices
        """

    @abstractmethod
    def get_slices(self) -> List[ISlice]:
        """
        Returns all slices registered with the actor.

        Returns:
            an array of slices
        """

    @abstractmethod
    def get_slice(self, *, slice_id: ID) -> ISlice:
        """
        Returns the slice with the given id.

        Args:
            slice_id: slice id
        Returns:
            the slice
        """

    @abstractmethod
    def register_slice(self, *, slice_object: ISlice):
        """
        Registers the slice with the actor. The slice must be a newly created one without a database record.
        If the slice is a recovered/previously unregistered one use re_register_slice(Slice) instead.

        Args:
            slice_object: slice
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def remove_slice(self, *, slice_object: ISlice):
        """
        Removes the specified slice. Purges slice-related state from the database.

        Args:
            slice_object: slice
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def remove_slice_by_slice_id(self, *, slice_id: ID):
        """
        Removes the specified slice. Purges slice-related state from the database.

        Args:
            slice_id: slice id
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def re_register_slice(self, *, slice_object: ISlice):
        """
        Re-registers the slice with the actor. The slice must already have a database record.

        Args:
            slice_object: slice
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def unregister_slice(self, *, slice_object: ISlice):
        """
        Unregisters the slice. Does not purge slice-related state from the database.

        Args:
            slice_object: slice
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def unregister_slice_by_slice_id(self, *, slice_id: ID):
        """
        Unregisters the slice. Does not purge slice-related state from the database.

        Args:
            slice_id: slice_id
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def close_slice_reservations(self, *, slice_id: ID):
        """
        Close slice reservations

        Args:
            slice_id: slice_id
        Raises:
            Exception in case of error
        """
